from typing import Union

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.messaging.shared_schemas import (
    MessagingServiceSecretsAWS_SES,
    MessagingServiceSecretsMailchimpTransactional,
    MessagingServiceSecretsMailgun,
    MessagingServiceSecretsTwilioEmail,
    MessagingServiceSecretsTwilioSMS,
)


class MessagingServiceSecretsMailchimpTransactionalDocs(
    MessagingServiceSecretsMailchimpTransactional,
    NoValidationSchema,
):
    """The secrets required to connect Mailchimp Transactional, for documentation"""


class MessagingSecretsMailgunDocs(MessagingServiceSecretsMailgun, NoValidationSchema):
    """The secrets required to connect to Mailgun, for documentation"""


class MessagingSecretsTwilioSMSDocs(
    MessagingServiceSecretsTwilioSMS, NoValidationSchema
):
    """The secrets required to connect to Twilio sms, for documentation"""


class MessagingSecretsTwilioEmailDocs(
    MessagingServiceSecretsTwilioEmail, NoValidationSchema
):
    """The secrets required to connect to Twilio email, for documentation"""


class MessagingServiceSecretsAWS_SESDocs(
    MessagingServiceSecretsAWS_SES, NoValidationSchema
):
    """The secrets required to connect to AWS SES, for documentation"""


possible_messaging_secrets = Union[
    MessagingSecretsMailgunDocs,
    MessagingSecretsTwilioSMSDocs,
    MessagingSecretsTwilioEmailDocs,
    MessagingServiceSecretsMailchimpTransactionalDocs,
    MessagingServiceSecretsAWS_SESDocs,
]



from __future__ import annotations

from datetime import datetime
from typing import Dict, Optional, Type, Union

from loguru import logger
from pydantic import ValidationError
from sqlalchemy import Boolean, Column, DateTime, Enum, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.common_exceptions import MessageDispatchException
from fides.api.db.base_class import Base, JSONTypeOverride
from fides.api.schemas.messaging.messaging import (
    EMAIL_MESSAGING_SERVICES,
    SMS_MESSAGING_SERVICES,
    MessagingConnectionTestStatus,
    MessagingMethod,
    MessagingServiceType,
)
from fides.api.schemas.messaging.shared_schemas import (
    SUPPORTED_MESSAGING_SERVICE_SECRETS,
    MessagingServiceSecretsAWS_SES,
    MessagingServiceSecretsMailchimpTransactional,
    MessagingServiceSecretsMailgun,
    MessagingServiceSecretsTwilioEmail,
    MessagingServiceSecretsTwilioSMS,
    PossibleMessagingSecrets,
)
from fides.api.util.logger import Pii
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy


def get_messaging_method(
    service_type: Optional[str],
) -> Optional[MessagingMethod]:
    """returns messaging method based on configured service type"""
    if service_type in EMAIL_MESSAGING_SERVICES:
        return MessagingMethod.EMAIL
    if service_type in SMS_MESSAGING_SERVICES:
        return MessagingMethod.SMS
    return None


def get_schema_for_secrets(
    service_type: MessagingServiceType,
    secrets: PossibleMessagingSecrets,
) -> SUPPORTED_MESSAGING_SERVICE_SECRETS:
    """
    Returns the secrets that pertain to `service_type` represented as a Pydantic schema
    for validation purposes.
    """
    messaging_secret_schema_type = Union[
        Type[MessagingServiceSecretsMailgun],
        Type[MessagingServiceSecretsTwilioSMS],
        Type[MessagingServiceSecretsTwilioEmail],
        Type[MessagingServiceSecretsMailchimpTransactional],
        Type[MessagingServiceSecretsAWS_SES],
    ]
    try:
        schema_mapping: Dict[MessagingServiceType, messaging_secret_schema_type] = {
            MessagingServiceType.mailgun: MessagingServiceSecretsMailgun,
            MessagingServiceType.twilio_text: MessagingServiceSecretsTwilioSMS,
            MessagingServiceType.twilio_email: MessagingServiceSecretsTwilioEmail,
            MessagingServiceType.mailchimp_transactional: MessagingServiceSecretsMailchimpTransactional,
            MessagingServiceType.aws_ses: MessagingServiceSecretsAWS_SES,
        }
        schema: messaging_secret_schema_type = schema_mapping[service_type]
    except KeyError:
        raise ValueError(
            f"`service_type` {service_type} has no supported `secrets` validation."
        )

    return schema.model_validate(secrets)


class MessagingConfig(Base):
    """The DB ORM model for MessagingConfig"""

    key = Column(String, index=True, unique=True, nullable=False)
    name = Column(String, unique=True, index=True)
    service_type = Column(
        Enum(MessagingServiceType), index=True, unique=True, nullable=False
    )
    details = Column(MutableDict.as_mutable(JSONB), nullable=True)
    last_test_timestamp = Column(DateTime(timezone=True))
    last_test_succeeded = Column(Boolean)
    secrets = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db

    @classmethod
    def get_configuration(cls, db: Session, service_type: str) -> MessagingConfig:
        """
        Fetches the configured MessagingConfig record by service type. Once fetched this function validates that
        the MessagingConfig is configured with secrets.
        """
        instance: Optional[MessagingConfig] = cls.get_by(
            db=db, field="service_type", value=service_type
        )
        if not instance:
            raise MessageDispatchException(
                f"No messaging config found for service_type {service_type}."
            )
        if not instance.secrets:
            logger.warning(
                "Messaging secrets not found for config with key: {}", instance.key
            )
            raise MessageDispatchException(
                f"Messaging secrets not found for config with key: {instance.key}"
            )
        return instance

    @classmethod
    def get_by_type(
        cls,
        db: Session,
        service_type: MessagingServiceType,
    ) -> Optional[MessagingConfig]:
        """
        Retrieve the messaging config of the given type
        """
        return db.query(cls).filter_by(service_type=service_type.value).first()

    def set_secrets(
        self,
        *,
        db: Session,
        messaging_secrets: PossibleMessagingSecrets,
    ) -> None:
        """Creates or updates secrets associated with a config id"""

        service_type = self.service_type
        if not service_type:
            raise ValueError(
                "This object must have a `service_type` to validate secrets."
            )

        try:
            get_schema_for_secrets(
                service_type=service_type,  # type: ignore
                secrets=messaging_secrets,
            )
        except (
            KeyError,
            ValidationError,
        ) as exc:
            logger.error("Error: {}", Pii(str(exc)))
            # We don't want to handle these explicitly here, only in the API view
            raise

        self.secrets = messaging_secrets
        self.save(db=db)

    @classmethod
    def get_active_default(cls, db: Session) -> Optional[MessagingConfig]:
        """
        Utility method to return the active default messaging configuration.

        This is determined by looking at the `notifications.notification_service_type`
        config property. We determine that config property's value by using
        the ConfigProxy, which resolves the value based on API-set or traditional
        config-set mechanisms.
        """
        active_default_messaging_type = ConfigProxy(
            db
        ).notifications.notification_service_type
        if not active_default_messaging_type:
            return None
        try:
            service_type = MessagingServiceType[active_default_messaging_type]
            return cls.get_by_type(db, service_type)
        except KeyError:
            raise ValueError(
                f"Unknown notification_service_type {active_default_messaging_type} configured"
            )

    def update_test_status(
        self, test_status: MessagingConnectionTestStatus, db: Session
    ) -> None:
        """
        Updates last_test_timestamp and last_test_succeeded after an attempt to send a test message.
        """
        self.last_test_timestamp = datetime.now()
        self.last_test_succeeded = (
            test_status == MessagingConnectionTestStatus.succeeded
        )
        self.save(db)


def default_messaging_config_name(service_type: str) -> str:
    """
    Utility function for consistency in generating default message config names.

    Returns a name to be used in a default messaging config for the given type.
    """
    return MessagingServiceType(service_type).human_readable


def default_messaging_config_key(service_type: str) -> str:
    """
    Utility function for consistency in generating default message config keys.

    Returns a key to be used in a default messaging config for the given type.
    """
    # Historically, when a key was not supplied, the service type string itself
    # was used as the key (e.g. "mailgun").  Reverting to that behaviour keeps
    # backward-compatibility with existing tests and API consumers that may rely
    # on this convention.
    return service_type.lower()



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, field_validator, model_validator


class RateLimitPeriod(str, Enum):
    """
    Defines the periods supported by rate limit config
    """

    second = "second"
    minute = "minute"
    hour = "hour"
    day = "day"


class RateLimit(BaseModel):
    """
    A config object which allows configuring rate limits for connectors
    """

    rate: int
    period: RateLimitPeriod
    custom_key: Optional[str] = None

    @field_validator("rate")
    @classmethod
    def rate_more_than_zero(cls, v: int) -> int:
        assert v > 0, "rate must be more than zero"
        return v


class RateLimitConfig(BaseModel):
    """
    A config object which allows configuring rate limits for connectors
    """

    limits: Optional[List[RateLimit]] = None
    enabled: Optional[bool] = True

    @model_validator(mode="after")
    def validate_all(self) -> "RateLimitConfig":
        limits: Optional[List[RateLimit]] = self.limits
        enabled: Optional[bool] = self.enabled

        if enabled:
            assert (
                limits and len(limits) > 0
            ), "limits must be set if rate limiter is enabled"
        if not enabled:
            assert not limits, "limits cannot be set if enabled is false"
        return self



"""
Contains the code that sets up the API.
"""

import os
import re
import sys
from datetime import datetime, timezone
from logging import WARNING
from time import perf_counter
from typing import AsyncGenerator, Callable, Optional
from urllib.parse import unquote

from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import FileResponse, HTMLResponse, JSONResponse
from fideslog.sdk.python.event import AnalyticsEvent
from loguru import logger
from pyinstrument import Profiler
from slowapi import _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded
from starlette.background import BackgroundTask
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY
from uvicorn import Config, Server

import fides
from fides.api.app_setup import (
    check_redis,
    create_fides_app,
    log_startup,
    run_database_startup,
)
from fides.api.common_exceptions import MalisciousUrlException
from fides.api.cryptography.identity_salt import get_identity_salt
from fides.api.middleware import handle_audit_log_resource
from fides.api.migrations.hash_migration_job import initiate_bcrypt_migration_task
from fides.api.migrations.post_upgrade_index_creation import (
    initiate_post_upgrade_index_creation,
)
from fides.api.schemas.analytics import Event, ExtraData

# pylint: disable=wildcard-import, unused-wildcard-import
from fides.api.service.privacy_request.email_batch_service import (
    initiate_scheduled_batch_email_send,
)
from fides.api.service.privacy_request.request_service import (
    initiate_async_tasks_status_polling,
    initiate_interrupted_task_requeue_poll,
    initiate_poll_for_exited_privacy_request_tasks,
    initiate_scheduled_dsr_data_removal,
)

# pylint: disable=wildcard-import, unused-wildcard-import
from fides.api.service.saas_request.override_implementations import *
from fides.api.tasks.scheduled.scheduler import async_scheduler, scheduler
from fides.api.ui import (
    get_admin_index_as_response,
    get_path_to_admin_ui_file,
    get_ui_file_map,
    match_route,
    path_is_in_ui_directory,
)
from fides.api.util.endpoint_utils import API_PREFIX
from fides.api.util.logger import _log_exception
from fides.api.util.rate_limit import safe_rate_limit_key
from fides.cli.utils import FIDES_ASCII_ART
from fides.config import CONFIG, check_required_webserver_config_values

IGNORED_AUDIT_LOG_RESOURCE_PATHS = {"/api/v1/login"}
NEXT_JS_CATCH_ALL_SEGMENTS_RE = r"^\[{1,2}\.\.\.\w+\]{1,2}"  # https://nextjs.org/docs/pages/building-your-application/routing/dynamic-routes#catch-all-segments

VERSION = fides.__version__


async def lifespan(wrapped_app: FastAPI) -> AsyncGenerator[None, None]:
    """Run all of the required setup steps for the webserver.

    **NOTE**: The order of operations here _is_ deliberate
    and must be maintained.
    """
    start_time = perf_counter()
    logger.info("Starting server setup...")

    if not CONFIG.dev_mode:
        sys.tracebacklimit = 0

    log_startup()

    await run_database_startup(wrapped_app)

    check_redis()

    if not scheduler.running:
        scheduler.start()
    if not async_scheduler.running:
        async_scheduler.start()

    # generate and/or cache the identity salt
    get_identity_salt()

    initiate_scheduled_batch_email_send()
    initiate_poll_for_exited_privacy_request_tasks()
    initiate_scheduled_dsr_data_removal()
    initiate_interrupted_task_requeue_poll()
    initiate_async_tasks_status_polling()
    initiate_bcrypt_migration_task()
    initiate_post_upgrade_index_creation()

    logger.debug("Sending startup analytics events...")
    # Avoid circular imports
    from fides.api.analytics import in_docker_container, send_analytics_event

    await send_analytics_event(
        AnalyticsEvent(
            docker=in_docker_container(),
            event=Event.server_start.value,
            event_created_at=datetime.now(tz=timezone.utc),
        )
    )

    # It's just a random bunch of strings when serialized
    if not CONFIG.logging.serialization:
        logger.info(FIDES_ASCII_ART)

    warn_root_user_enabled()

    logger.info("Fides startup complete! v{}", VERSION)
    startup_time = round(perf_counter() - start_time, 3)
    logger.info("Server setup completed in {} seconds", startup_time)
    yield  # All of this happens before the webserver comes up


app = create_fides_app(lifespan=lifespan)  # type: ignore


if CONFIG.dev_mode:

    @app.middleware("http")
    async def profile_request(request: Request, call_next: Callable) -> Response:
        profiling = request.headers.get("profile-request", False)
        if profiling:
            profiler = Profiler(interval=0.001, async_mode="enabled")
            profiler.start()
            await call_next(request)
            profiler.stop()
            logger.debug("Request Profiled!")
            return HTMLResponse(profiler.output_text(timeline=True, show_all=True))

        return await call_next(request)


@app.middleware("http")
async def dispatch_log_request(request: Request, call_next: Callable) -> Response:
    """
    HTTP Middleware that logs analytics events for each call to Fides endpoints.
    :param request: Request to Fides api
    :param call_next: Callable api endpoint
    :return: Response
    """

    # Only log analytics events for requests that are for API endpoints (i.e. /api/...)
    path = request.url.path
    if (not path.startswith(API_PREFIX)) or (path.endswith("/health")):
        return await call_next(request)

    fides_source: Optional[str] = request.headers.get("X-Fides-Source")
    now: datetime = datetime.now(tz=timezone.utc)
    endpoint = f"{request.method}: {request.url}"

    try:
        response = await call_next(request)
        # HTTPExceptions are considered a handled err by default so are not thrown here.
        # Accepted workaround is to inspect status code of response.
        # More context- https://github.com/tiangolo/fastapi/issues/1840
        response.background = BackgroundTask(
            prepare_and_log_request,
            endpoint,
            request.url.hostname,
            response.status_code,
            now,
            fides_source,
            "HTTPException" if response.status_code >= 400 else None,
        )
        return response

    except Exception as e:
        await prepare_and_log_request(
            endpoint, request.url.hostname, 500, now, fides_source, e.__class__.__name__
        )
        _log_exception(e, CONFIG.dev_mode)
        raise


async def prepare_and_log_request(
    endpoint: str,
    hostname: Optional[str],
    status_code: int,
    event_created_at: datetime,
    fides_source: Optional[str],
    error_class: Optional[str],
) -> None:
    """
    Prepares and sends analytics event provided the user is not opted out of analytics.
    """
    # Avoid circular imports
    from fides.api.analytics import (
        accessed_through_local_host,
        in_docker_container,
        send_analytics_event,
    )

    # this check prevents AnalyticsEvent from being called with invalid endpoint during unit tests
    if CONFIG.user.analytics_opt_out:
        return
    await send_analytics_event(
        AnalyticsEvent(
            docker=in_docker_container(),
            event=Event.endpoint_call.value,
            event_created_at=event_created_at,
            local_host=accessed_through_local_host(hostname),
            endpoint=endpoint,
            status_code=status_code,
            error=error_class or None,
            extra_data=(
                {ExtraData.fides_source.value: fides_source} if fides_source else None
            ),
        )
    )


@app.middleware("http")
async def log_request(request: Request, call_next: Callable) -> Response:
    """Log basic information about every request handled by the server."""
    start = datetime.now()

    # If the request fails, we still want to log it
    try:
        response = await call_next(request)
    except Exception as e:  # pylint: disable=bare-except
        logger.exception(f"Unhandled exception processing request: '{e}'")
        response = Response(status_code=500)

    handler_time = datetime.now() - start

    # Take the total time in seconds and convert it to milliseconds, rounding to 3 decimal places
    total_time = round(handler_time.total_seconds() * 1000, 3)
    logger.bind(
        method=request.method,
        status_code=response.status_code,
        handler_time=f"{total_time}ms",
        path=request.url.path,
        fides_client=request.headers.get("Fides-Client", "unknown"),
    ).info("Request received")
    return response


# Configure the static file paths last since otherwise it will take over all paths
@app.get("/", tags=["Default"])
def read_index() -> Response:
    """
    Return an index.html at the root path
    """

    return get_admin_index_as_response()


def sanitise_url_path(path: str) -> str:
    """Returns a URL path that does not contain any ../ or //"""
    path = unquote(path)
    path = os.path.normpath(path)

    for token in path.split("/"):
        if ".." in token and not re.search(NEXT_JS_CATCH_ALL_SEGMENTS_RE, token):
            logger.warning(
                f"Potentially dangerous use of URL hierarchy in path: {path}"
            )
            raise MalisciousUrlException()
    return path


@app.get("/{catchall:path}", response_class=Response, tags=["Default"])
def read_other_paths(request: Request) -> Response:
    """
    Return related frontend files. Adapted from https://github.com/tiangolo/fastapi/issues/130
    """
    # check first if requested file exists (for frontend assets)
    path = request.path_params["catchall"]
    logger.debug(f"Catch all path detected: {path}")
    try:
        path = sanitise_url_path(path)
    except MalisciousUrlException:
        # if a maliscious URL is detected, route the user to the index
        return get_admin_index_as_response()

    # search for matching route in package (i.e. /dataset)
    ui_file = match_route(get_ui_file_map(), path)

    # if not, check if the requested file is an asset (i.e. /_next/static/...)
    if not ui_file:
        ui_file = get_path_to_admin_ui_file(path)

    # Serve up the file as long as it is within the UI directory
    if ui_file and ui_file.is_file():
        if not path_is_in_ui_directory(ui_file):
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Item not found"
            )
        logger.debug(
            "catchall request path '{}' matched static admin UI file: {}",
            path,
            ui_file,
        )
        return FileResponse(ui_file)

    # raise 404 for anything that should be backend endpoint but we can't find it
    if path.startswith(API_PREFIX[1:]):
        logger.debug(
            "catchall request path '{}' matched an invalid API route, return 404",
            path,
        )
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Item not found"
        )

    # otherwise return the index
    logger.debug(
        "catchall request path '{}' did not match any admin UI routes, return generic admin UI index",
        path,
    )
    return get_admin_index_as_response()


def warn_root_user_enabled() -> None:
    """
    Log a startup warning if root user is enabled.
    Extracted as a function because this may need to be done in multiple places,
    depending on how the server is started.
    """
    if CONFIG.security.root_username or CONFIG.security.oauth_root_client_id:
        logger.warning(
            "Root Username & Password are configured and can be used to login as a root user. If unexpected, review security settings (FIDES__SECURITY__ROOT_USERNAME and FIDES__SECURITY__ROOT_PASSWORD)"
        )


def start_webserver(port: int = 8080) -> None:
    """Run the webserver."""
    check_required_webserver_config_values(config=CONFIG)
    server = Server(Config(app, host="0.0.0.0", port=port, log_level=WARNING))

    logger.info(
        "Starting webserver - Host: {}, Port: {}, Log Level: {}",
        server.config.host,
        server.config.port,
        server.config.log_level,
    )

    warn_root_user_enabled()

    server.run()


@app.middleware("http")
async def action_to_audit_log(
    request: Request,
    call_next: Callable,
) -> Response:
    """Log basic information about every non-GET request handled by the server."""

    if (
        request.method != "GET"
        and request.scope["path"] not in IGNORED_AUDIT_LOG_RESOURCE_PATHS
        and CONFIG.security.enable_audit_log_resource_middleware
    ):
        try:
            await handle_audit_log_resource(request)
        except Exception as exc:
            logger.debug(exc)
    return await call_next(request)


@app.exception_handler(RequestValidationError)
async def request_validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    """Prevents field input and pydantic error message URL from being returned in the response
    For example, if someone is creating a user and the request fails, this prevents the user's
    password from being returned in the error message
    """
    return JSONResponse(
        status_code=HTTP_422_UNPROCESSABLE_ENTITY,
        content={
            "detail": jsonable_encoder(exc.errors(), exclude={"input", "url", "ctx"})
        },
    )


@app.exception_handler(RateLimitExceeded)
async def rate_limit_handler(request: Request, exc: RateLimitExceeded) -> Response:
    """Log rate limit violations and delegate to default handler."""
    client_ip = safe_rate_limit_key(
        request
    )  # non exception-raising, falls back to source IP

    # Log the rate limit event
    logger.warning(
        "Rate limit exceeded - IP: %s, Path: %s, Method: %s",
        client_ip,
        request.url.path,
        request.method,
    )

    # Use the default handler to generate the proper response
    return _rate_limit_exceeded_handler(request, exc)



# pylint: disable=too-many-lines
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import dask
from dask import delayed  # type: ignore[attr-defined]
from dask.core import getcycle
from dask.threaded import get
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import TraversalError
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
    CollectionAddress,
)
from fides.api.graph.graph import DatasetGraph
from fides.api.graph.traversal import (
    Traversal,
    TraversalNode,
    log_traversal_error_and_update_privacy_request,
)
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.sql_models import System  # type: ignore[attr-defined]
from fides.api.schemas.policy import ActionType
from fides.api.task.graph_task import EMPTY_REQUEST_TASK, GraphTask
from fides.api.task.task_resources import TaskResources
from fides.api.util.collection_util import Row

# These are deprecated DSR 2.0 functions that support running DSR's in sequence with Dask in-memory
# Supported for a limited time.

dask.config.set(scheduler="threads")


def update_mapping_from_cache(
    dsk: Dict[CollectionAddress, Tuple[Any, ...]],
    resources: TaskResources,
    start_fn: Callable,
) -> None:
    """When resuming a privacy request from a paused or failed state, update the `dsk` dictionary with results we've
    already obtained from a previous run. Remove upstream dependencies for these nodes, and just return the data we've
    already retrieved, rather than visiting them again.

    If there's no cached data, the dsk dictionary won't change.
    """

    cached_results: Dict[str, Optional[List[Row]]] = resources.get_all_cached_objects()

    for collection_name in cached_results:
        dsk[CollectionAddress.from_string(collection_name)] = (
            start_fn(cached_results[collection_name]),
        )


def format_data_use_map_for_caching(
    connection_key_mapping: Dict[CollectionAddress, str],
    connection_configs: List[ConnectionConfig],
) -> Dict[str, Set[str]]:
    """
    Create a map of `Collection`s mapped to their associated `DataUse`s
    to be stored in the cache. This is done before request execution, so that we
    maintain the _original_ state of the graph as it's used for request execution.
    The graph is subject to change "from underneath" the request execution runtime,
    but we want to avoid picking up those changes in our data use map.

    `DataUse`s are associated with a `Collection` by means of the `System`
    that's linked to a `Collection`'s `Connection` definition.

    Example:
    {
       <collection1>: {"data_use_1", "data_use_2"},
       <collection2>: {"data_use_1"},
    }
    """
    resp: Dict[str, Set[str]] = {}
    connection_config_mapping: Dict[str, ConnectionConfig] = {
        connection_config.key: connection_config
        for connection_config in connection_configs
    }
    for collection_addr, connection_key in connection_key_mapping.items():
        connection_config = connection_config_mapping.get(connection_key, None)
        if not connection_config or not connection_config.system:
            resp[collection_addr.value] = set()
            continue
        data_uses: Set[str] = System.get_data_uses(
            [connection_config.system], include_parents=False
        )
        resp[collection_addr.value] = data_uses

    return resp


def start_function(seed: List[Dict[str, Any]]) -> Callable[[], List[Dict[str, Any]]]:
    """Return a function for collections with no upstream dependencies, that just start
    with seed data.

    This is used for root nodes or previously-visited nodes on restart."""

    def g() -> List[Dict[str, Any]]:
        return seed

    return g


def run_access_request_deprecated(
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
) -> Dict[str, List[Row]]:
    """Deprecated: Run the access request sequentially in-memory using Dask"""
    try:
        traversal: Traversal = Traversal(graph, identity, policy=policy)
        # Add execution logs for skipped nodes
        if traversal.skipped_nodes:
            logger.warning(
                "Some nodes were skipped, the identities provided were not sufficient to reach them"
            )

            for node_address, skip_message in traversal.skipped_nodes.items():
                logger.debug(skip_message)
                privacy_request.add_skipped_execution_log(
                    session,
                    connection_key=None,
                    dataset_name="Dataset traversal",
                    collection_name=node_address.replace(":", "."),
                    message=skip_message,
                    action_type=ActionType.access,
                )
        # Or log success if all collections are reachable
        else:
            privacy_request.add_success_execution_log(
                session,
                connection_key=None,
                dataset_name="Dataset traversal",
                collection_name=None,
                message=f"Traversal successful for privacy request: {privacy_request.id}",
                action_type=ActionType.access,
            )

    except TraversalError as err:
        log_traversal_error_and_update_privacy_request(privacy_request, session, err)
        raise err

    with TaskResources(
        privacy_request, policy, connection_configs, EMPTY_REQUEST_TASK, session
    ) as resources:

        def collect_tasks_fn(
            tn: TraversalNode, data: Dict[CollectionAddress, GraphTask]
        ) -> None:
            """Run the traversal, as an action creating a GraphTask for each traversal_node."""
            if not tn.is_root_node():
                # Mock a RequestTask object in memory to share code with DSR 3.0
                resources.privacy_request_task = tn.to_mock_request_task()
                data[tn.address] = GraphTask(resources)

        def termination_fn(
            *dependent_values: List[Row],
        ) -> Dict[str, Optional[List[Row]]]:
            """A termination function that just returns its inputs mapped to their source addresses.
            This needs to wait for all dependent keys because this is how dask is informed to wait for
            all terminating addresses before calling this."""

            return resources.get_all_cached_objects()

        env: Dict[CollectionAddress, GraphTask] = {}
        end_nodes: List[CollectionAddress] = traversal.traverse(env, collect_tasks_fn)

        dsk: Dict[CollectionAddress, Tuple[Any, ...]] = {
            k: (t.access_request, *t.execution_node.input_keys) for k, t in env.items()
        }
        dsk[ROOT_COLLECTION_ADDRESS] = (start_function([traversal.seed_data]),)
        dsk[TERMINATOR_ADDRESS] = (termination_fn, *end_nodes)
        update_mapping_from_cache(dsk, resources, start_function)

        # cache a map of collections -> data uses for the output package of access requests
        # this is cached here before request execution, since this is the state of the
        # graph used for request execution. the graph could change _during_ request execution,
        # but we don't want those changes in our data use map.
        privacy_request.cache_data_use_map(
            format_data_use_map_for_caching(
                {
                    coll_address: gt.execution_node.connection_key
                    for (coll_address, gt) in env.items()
                },
                connection_configs,
            )
        )

        v = delayed(get(dsk, TERMINATOR_ADDRESS, num_workers=1))
        return v.compute()


def update_erasure_mapping_from_cache(
    dsk: Dict[CollectionAddress, Union[Tuple[Any, ...], int]], resources: TaskResources
) -> None:
    """On pause or restart from failure, update the dsk graph to skip running erasures on collections
    we've already visited. Instead, just return the previous count of rows affected.

    If there's no cached data, the dsk dictionary won't change.
    """
    cached_erasures: Dict[str, int] = resources.get_all_cached_erasures()

    for collection_name in cached_erasures:
        dsk[CollectionAddress.from_string(collection_name)] = cached_erasures[
            collection_name
        ]


def run_erasure_request_deprecated(  # pylint: disable = too-many-arguments
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    access_request_data: Dict[str, List[Row]],
    session: Session,
) -> Dict[str, int]:
    """Deprecated: Run an erasure request sequentially in-memory using Dask"""
    traversal: Traversal = Traversal(graph, identity, policy=policy)
    with TaskResources(
        privacy_request, policy, connection_configs, EMPTY_REQUEST_TASK, session
    ) as resources:

        def collect_tasks_fn(
            tn: TraversalNode, data: Dict[CollectionAddress, GraphTask]
        ) -> None:
            """Run the traversal, as an action creating a GraphTask for each traversal_node."""
            if not tn.is_root_node():
                # Mock a RequestTask object in memory to share code with DSR 3.0
                resources.privacy_request_task = tn.to_mock_request_task()
                data[tn.address] = GraphTask(resources)

        env: Dict[CollectionAddress, GraphTask] = {}
        # Modifies env in place
        traversal.traverse(env, collect_tasks_fn)
        erasure_end_nodes = list(graph.nodes.keys())

        def termination_fn(*dependent_values: int) -> Dict[str, int]:
            """
            The erasure order can be affected in a way that not every node is directly linked
            to the termination node. This means that we can't just aggregate the inputs directly,
            we must read the erasure results from the cache.
            """
            return resources.get_all_cached_erasures()

        access_request_data[ROOT_COLLECTION_ADDRESS.value] = [identity]

        # Build the dask task graph for erasure, ensuring we also pass along the
        # upstream access data that may be required by the connector (e.g. BigQuery
        # delete statements).  We accomplish this by partially applying the
        # `inputs` kwarg on each task's `erasure_request` method.  The resulting
        # callable accepts the original positional arguments expected by Dask.

        dsk: Dict[CollectionAddress, Any] = {}
        for k, t in env.items():
            # Collect upstream access data in the same order as the input keys
            upstream_access_data: List[List[Row]] = [
                access_request_data.get(str(input_key), [])
                for input_key in t.execution_node.input_keys
            ]

            # Bind the `inputs` keyword argument so it is supplied when the task executes
            erasure_fn_with_inputs = partial(
                t.erasure_request, inputs=upstream_access_data
            )

            # Build the task tuple: (callable, retrieved_data, *prereqs)
            dsk[k] = (
                erasure_fn_with_inputs,
                access_request_data.get(
                    str(k), []
                ),  # Data retrieved for this collection
                *_evaluate_erasure_dependencies(t, erasure_end_nodes),
            )

        # root node returns 0 to be consistent with the output of the other erasure tasks
        dsk[ROOT_COLLECTION_ADDRESS] = 0
        # terminator function reads and returns the cached erasure results for the entire erasure traversal
        dsk[TERMINATOR_ADDRESS] = (termination_fn, *erasure_end_nodes)
        update_erasure_mapping_from_cache(dsk, resources)

        # using an existing function from dask.core to detect cycles in the generated graph
        collection_cycle = getcycle(dsk, None)
        if collection_cycle:
            logger.error(
                "TraversalError encountered for privacy request {}. Error: The values for the `erase_after` fields caused a cycle in the following collections {}",
                privacy_request.id,
                collection_cycle,
            )
            privacy_request.add_error_execution_log(
                db=session,
                connection_key=None,
                collection_name=None,
                dataset_name=None,
                message=f"The values for the `erase_after` fields caused a cycle in the following collections {collection_cycle}",
                action_type=ActionType.erasure,
            )
            privacy_request.error_processing(session)
            raise TraversalError(
                f"The values for the `erase_after` fields caused a cycle in the following collections {collection_cycle}"
            )

        v = delayed(get(dsk, TERMINATOR_ADDRESS, num_workers=1))
        return v.compute()


def _evaluate_erasure_dependencies(
    t: GraphTask, end_nodes: List[CollectionAddress]
) -> Set[CollectionAddress]:
    """
    Return a set of collection addresses corresponding to collections that need
    to be erased before the given task. Remove the dependent collection addresses
    from `end_nodes` so they can be executed in the correct order. If a task does
    not have any dependencies it is linked directly to the root node
    """
    erase_after = t.execution_node.collection.erase_after
    for collection in erase_after:
        if collection in end_nodes:
            # end_node list is modified in place
            end_nodes.remove(collection)
    # this task will execute after the collections in `erase_after` or
    # execute at the beginning by linking it to the root node
    return erase_after if len(erase_after) else {ROOT_COLLECTION_ADDRESS}


def run_consent_request_deprecated(  # pylint: disable = too-many-arguments
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
) -> Dict[str, bool]:
    """Run a consent request

    The graph built is very simple: there are no relationships between the nodes, every node has
    identity data input and every node outputs whether the consent request succeeded.

    The DatasetGraph passed in is expected to have one Node per Dataset.  That Node is expected to carry out requests
    for the Dataset as a whole.
    """
    with TaskResources(
        privacy_request, policy, connection_configs, EMPTY_REQUEST_TASK, session
    ) as resources:
        graph_keys: List[CollectionAddress] = list(graph.nodes.keys())
        dsk: Dict[CollectionAddress, Any] = {}

        for col_address, node in graph.nodes.items():
            traversal_node = TraversalNode(node)
            # Mock a RequestTask object in memory to share code with DSR 3.0
            resources.privacy_request_task = traversal_node.to_mock_request_task()
            task = GraphTask(resources)
            dsk[col_address] = (task.consent_request, identity)

        def termination_fn(*dependent_values: bool) -> Tuple[bool, ...]:
            """The dependent_values here is an bool output from each task feeding in, where
            each task reports the output of 'task.consent_request(identity_data)', which is whether the
            consent request succeeded

            The termination function just returns this tuple of booleans."""
            return dependent_values

        # terminator function waits for all keys
        dsk[TERMINATOR_ADDRESS] = (termination_fn, *graph_keys)

        v = delayed(get(dsk, TERMINATOR_ADDRESS, num_workers=1))

        update_successes: Tuple[bool, ...] = v.compute()
        # we combine the output of the termination function with the input keys to provide
        # a map of {collection_name: whether consent request succeeded}:
        consent_update_map: Dict[str, bool] = dict(
            zip([coll.value for coll in graph_keys], update_successes)
        )

        return consent_update_map



from typing import Any, Optional

from sqlalchemy import Boolean, Column, String, insert, select
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr

from fides.api.db.base_class import Base


class WebMonitorGroupJob(Base):
    """SQL model for storing data returned from a web monitor group job"""

    @declared_attr
    def __tablename__(self) -> str:
        return "web_monitor_group_job"

    # The id for the web monitor group on the web monitor service
    group_id = Column(String, nullable=False, unique=True)

    # The raw experiences data obtained for that group from the web monitor service
    raw_experiences_data = Column(JSONB, nullable=True)
    # The processed experiences data, used to calculate an StagedResource's consent status
    processed_experiences_data = Column(JSONB, nullable=True)

    # Whether this data is from a test_website monitor rather than a real one
    is_test_run = Column(Boolean, nullable=False, default=False)

    @classmethod
    async def create_async(
        cls,
        async_session: AsyncSession,
        *,
        group_id: str,
        is_test_run: Optional[bool] = False,
        raw_experiences_data: Optional[dict[str, Any]] = None,
        processed_experiences_data: Optional[dict[str, Any]] = None,
    ) -> "WebMonitorGroupJob":
        """Create a new WebMonitorGroupJob"""
        values = {
            "group_id": group_id,
            "is_test_run": is_test_run,
            "raw_experiences_data": raw_experiences_data,
            "processed_experiences_data": processed_experiences_data,
        }
        stmt = insert(cls.__table__).values(**values).returning(*cls.__table__.columns)
        result = await async_session.execute(stmt)
        row = result.first()
        return cls(**dict(row))

    @classmethod
    async def get_by_group_id_async(
        cls, async_session: AsyncSession, group_id: str
    ) -> Optional["WebMonitorGroupJob"]:
        """Get a WebMonitorGroupJob by group_id"""
        stmt = select(*cls.__table__.columns).where(
            cls.__table__.c.group_id == group_id
        )
        result = await async_session.execute(stmt)
        row = result.first()
        return cls(**dict(row)) if row else None



from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from fides.api.api.deps import get_config, get_config_proxy, get_db
from fides.api.db.ctl_session import get_async_db
from fides.config import FidesConfig
from fides.config.config_proxy import ConfigProxy
from fides.service.dataset.dataset_config_service import DatasetConfigService
from fides.service.dataset.dataset_service import DatasetService
from fides.service.event_audit_service import EventAuditService
from fides.service.messaging.messaging_service import MessagingService
from fides.service.privacy_request.privacy_request_service import PrivacyRequestService
from fides.service.system.system_service import SystemService
from fides.service.taxonomy.taxonomy_service import TaxonomyService
from fides.service.user.user_service import UserService


def get_messaging_service(
    db: Session = Depends(get_db),
    config: FidesConfig = Depends(get_config),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
) -> MessagingService:
    return MessagingService(db, config, config_proxy)


def get_privacy_request_service(
    db: Session = Depends(get_db),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
    messaging_service: MessagingService = Depends(get_messaging_service),
) -> PrivacyRequestService:
    return PrivacyRequestService(db, config_proxy, messaging_service)


def get_dataset_service(db: Session = Depends(get_db)) -> DatasetService:
    return DatasetService(db)


def get_dataset_config_service(db: Session = Depends(get_db)) -> DatasetConfigService:
    return DatasetConfigService(db)


def get_user_service(
    db: Session = Depends(get_db),
    config: FidesConfig = Depends(get_config),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
) -> UserService:
    return UserService(db, config, config_proxy)


def get_event_audit_service(db: Session = Depends(get_db)) -> EventAuditService:
    return EventAuditService(db)


def get_taxonomy_service(
    db: Session = Depends(get_db),
    event_audit_service: EventAuditService = Depends(get_event_audit_service),
) -> TaxonomyService:
    return TaxonomyService(db, event_audit_service)


def get_system_service(db: AsyncSession = Depends(get_async_db)) -> SystemService:
    return SystemService(db)



import numbers
import operator as py_operator

from fides.api.task.conditional_dependencies.schemas import GroupOperator, Operator

# Define operator methods for validation
#
# None Value Handling:
# - Basic operators (eq, neq, exists, not_exists) handle None naturally
#   * Note: eq and neq use Python's built-in behavior, so True == 1 returns True
# - Numeric operators return False for None values (can't compare None with numbers)
#   * Note: Boolean values are excluded from numeric comparisons (True < 10 returns False)
# - String operators return False for None values (can't perform string operations on None)
# - List operators handle None naturally using Python's built-in behavior:
#   * None in [None] returns True
#   * None not in [None] returns False
#   * None in [] returns False
#   * None not in [] returns True
#   * This allows None to be a valid list element for membership testing
OPERATOR_METHODS = {
    # Basic operators - handle None naturally using Python's built-in behavior
    Operator.exists: lambda a, _: a is not None,
    Operator.not_exists: lambda a, _: a is None,
    Operator.eq: py_operator.eq,
    # None == None returns True, None == "anything" returns False
    # None != None returns False, None != "anything" returns True
    Operator.neq: py_operator.ne,
    # Numeric comparison operators - return False for None or non-numeric types
    # Note: Boolean values are excluded as they are not considered numeric for comparisons
    # This differs from basic comparison operators (eq, neq) which use Python's built-in behavior
    Operator.lt: lambda a, b: (
        a < b
        if a is not None and isinstance(a, numbers.Number) and not isinstance(a, bool)
        else False
    ),
    Operator.lte: lambda a, b: (
        a <= b
        if a is not None and isinstance(a, numbers.Number) and not isinstance(a, bool)
        else False
    ),
    Operator.gt: lambda a, b: (
        a > b
        if a is not None and isinstance(a, numbers.Number) and not isinstance(a, bool)
        else False
    ),
    Operator.gte: lambda a, b: (
        a >= b
        if a is not None and isinstance(a, numbers.Number) and not isinstance(a, bool)
        else False
    ),
    Operator.list_contains: lambda a, b: (
        # If user provides a list, check if column value is in it
        # Note: Handles None values naturally using Python's built-in behavior
        a in b
        if isinstance(b, list)
        # If column value is a list, check if user value is in it
        else b in a if isinstance(a, list) else False
    ),
    Operator.not_in_list: lambda a, b: (
        # If user provides a list, check if column value is NOT in it
        # Note: Handles None values naturally using Python's built-in behavior
        a not in b
        if isinstance(b, list)
        # If column value is a list, check if user value is NOT in it
        else b not in a if isinstance(a, list) else False
    ),
    Operator.list_intersects: lambda a, b: (
        # Check if there are any common elements between the lists
        # Note: Returns False for None values because both values must be lists
        bool(set(a) & set(b))
        if isinstance(a, list) and isinstance(b, list)
        else False
    ),
    Operator.list_subset: lambda a, b: (
        # Check if column list is a subset of user's list
        # Note: Returns False for None values because both values must be lists
        set(a).issubset(set(b))
        if isinstance(a, list) and isinstance(b, list)
        else False
    ),
    Operator.list_superset: lambda a, b: (
        # Check if column list is a superset of user's list
        # Note: Returns False for None values because both values must be lists
        set(a).issuperset(set(b))
        if isinstance(a, list) and isinstance(b, list)
        else False
    ),
    Operator.list_disjoint: lambda a, b: (
        # Check if lists have no common elements
        # Note: Returns False for None values because both values must be lists
        set(a).isdisjoint(set(b))
        if isinstance(a, list) and isinstance(b, list)
        else False
    ),
    Operator.starts_with: lambda a, b: (
        a.startswith(b) if isinstance(a, str) and isinstance(b, str) else False
    ),
    Operator.ends_with: lambda a, b: (
        a.endswith(b) if isinstance(a, str) and isinstance(b, str) else False
    ),
    Operator.contains: lambda a, b: (
        b in a if isinstance(a, str) and isinstance(b, str) else False
    ),
}

# Define logical operators for group conditions
LOGICAL_OPERATORS = {
    GroupOperator.and_: all,
    GroupOperator.or_: any,
}

# Common operators that work with most data types
COMMON_OPERATORS = {
    Operator.eq,
    Operator.neq,
    Operator.exists,
    Operator.not_exists,
}

# Numeric comparison operators
NUMERIC_OPERATORS = {Operator.lt, Operator.lte, Operator.gt, Operator.gte}

# String-specific operators
STRING_OPERATORS = {
    Operator.contains,
    Operator.starts_with,
    Operator.ends_with,
}

# List operations that work with all data types
LIST_OPERATORS = {
    Operator.list_contains,  # Element in list
    Operator.not_in_list,  # Element not in list
    Operator.list_intersects,  # Any common elements
    Operator.list_subset,  # Column list ⊆ user list
    Operator.list_superset,  # Column list ⊇ user list
    Operator.list_disjoint,  # No common elements
}

# Define data type compatibility with operators
data_type_operator_compatibility = {
    "integer": {*COMMON_OPERATORS, *NUMERIC_OPERATORS, *LIST_OPERATORS},
    "float": {*COMMON_OPERATORS, *NUMERIC_OPERATORS, *LIST_OPERATORS},
    "double": {*COMMON_OPERATORS, *NUMERIC_OPERATORS, *LIST_OPERATORS},
    "long": {*COMMON_OPERATORS, *NUMERIC_OPERATORS, *LIST_OPERATORS},
    "boolean": {*COMMON_OPERATORS, *LIST_OPERATORS},
    "string": {*COMMON_OPERATORS, *STRING_OPERATORS, *LIST_OPERATORS},
    "text": {
        *COMMON_OPERATORS,
        *LIST_OPERATORS,
    },  # Form input - no string search operations
    "array": {*COMMON_OPERATORS, *LIST_OPERATORS},
    "object": {*COMMON_OPERATORS, *LIST_OPERATORS},
    "object_id": {*COMMON_OPERATORS, *LIST_OPERATORS},
    "no_op": {*COMMON_OPERATORS, *LIST_OPERATORS},
}



"""Add notifications for failed DSR requests

Revision ID: 28108b17a99c
Revises: 8f1a19465239
Create Date: 2022-11-07 23:56:14.066374

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "28108b17a99c"
down_revision = "8f1a19465239"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "privacyrequestnotifications",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("email", sa.String(), nullable=False),
        sa.Column("notify_after_failures", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyrequestnotifications_id"),
        "privacyrequestnotifications",
        ["id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_privacyrequestnotifications_id"),
        table_name="privacyrequestnotifications",
    )
    op.drop_table("privacyrequestnotifications")
    # ### end Alembic commands ###



"""add finalized fields to privacy request

Revision ID: a7065df4dcf1
Revises: 7e9a2b52f498
Create Date: 2025-07-01 14:07:36.779437

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a7065df4dcf1"
down_revision = "7e9a2b52f498"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "privacyrequest",
        sa.Column("finalized_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.add_column(
        "privacyrequest", sa.Column("finalized_by", sa.String(), nullable=True)
    )
    op.create_foreign_key(
        "privacyrequest_fidesuser_id_fkey",
        "privacyrequest",
        "fidesuser",
        ["finalized_by"],
        ["id"],
        ondelete="SET NULL",
    )
    op.execute(
        "alter type privacyrequeststatus add value 'requires_manual_finalization'"
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "privacyrequest_fidesuser_id_fkey", "privacyrequest", type_="foreignkey"
    )
    op.drop_column("privacyrequest", "finalized_by")
    op.drop_column("privacyrequest", "finalized_at")

    op.execute(
        "delete from privacyrequest where status in ('requires_manual_finalization')"
    )

    op.execute("alter type privacyrequeststatus rename to privacyrequeststatus_old")
    op.execute(
        "create type privacyrequeststatus as enum('identity_unverified', 'requires_input', 'pending', 'in_processing', 'complete', 'pending', 'error', 'paused', 'approved', 'denied', 'canceled', 'awaiting_email_send')"
    )
    op.execute(
        (
            "alter table privacyrequest alter column status type privacyrequeststatus using "
            "status::text::privacyrequeststatus"
        )
    )
    op.execute("drop type privacyrequeststatus_old")
    # ### end Alembic commands ###



import itertools
from typing import Any, Dict, Generator, List, Optional

from boto3 import Session
from boto3.dynamodb.types import TypeDeserializer
from botocore.exceptions import ClientError
from loguru import logger

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.connection_configuration.connection_secrets_dynamodb import (
    DynamoDBSchema,
)
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.dynamodb_query_config import (
    DynamoDBQueryConfig,
)
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.aws_util import get_aws_session
from fides.api.util.collection_util import Row
from fides.api.util.logger import Pii
from fides.connectors.models import (
    ConnectorAuthFailureException,
    ConnectorFailureException,
)


class DynamoDBConnector(BaseConnector[Any]):  # type: ignore
    """AWS DynamoDB Connector"""

    def build_uri(self) -> None:
        """Not used for this type"""

    def close(self) -> None:
        """Not used for this type"""

    def create_client(self) -> Any:  # type: ignore
        """Returns a client for a DynamoDB instance"""
        config = DynamoDBSchema(**self.configuration.secrets or {})
        try:
            aws_session: Session = get_aws_session(
                config.auth_method.value,  # pylint: disable=no-member
                config.model_dump(),  # type: ignore[arg-type]
                config.aws_assume_role_arn,
            )

            return aws_session.client("dynamodb")
        except ValueError:
            raise ConnectionException("Value Error connecting to AWS DynamoDB.")

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """Query wrapper corresponding to the input traversal_node."""
        client = self.client()
        try:
            describe_table = client.describe_table(TableName=node.address.collection)
            for key in describe_table["Table"]["KeySchema"]:
                if key["KeyType"] == "HASH":
                    hash_key = key["AttributeName"]
                    break
            for key in describe_table["Table"]["AttributeDefinitions"]:
                if key["AttributeName"] == hash_key:
                    attribute_definitions = [key]
                    break
        except ClientError as error:
            raise ConnectorFailureException(error.response["Error"]["Message"])

        return DynamoDBQueryConfig(node, attribute_definitions)

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Connects to AWS DynamoDB and lists tables to validate credentials.
        """
        logger.info("Starting test connection to {}", self.configuration.key)
        client = self.client()
        try:
            client.list_tables()
        except ClientError as error:
            if error.response["Error"]["Code"] in [
                "InvalidClientTokenId",
                "SignatureDoesNotMatch",
            ]:
                raise ConnectorAuthFailureException(error.response["Error"]["Message"])
            raise ConnectorFailureException(error.response["Error"]["Message"])

        return ConnectionTestStatus.succeeded

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """
        Retrieve DynamoDB data.
        In the case of complex objects, returns multiple rows
        as the product of options to query against.
        """
        deserializer = TypeDeserializer()
        collection_name = node.address.collection
        client = self.client()
        try:
            results = []
            query_config = self.query_config(node)
            for attribute_definition in query_config.attribute_definitions:  # type: ignore
                attribute_name = attribute_definition["AttributeName"]
                for identifier in input_data.get(attribute_name, []):
                    selected_input_data = input_data
                    selected_input_data[attribute_name] = [identifier]
                    query_param = query_config.generate_query(
                        selected_input_data, policy
                    )
                    if query_param is None:
                        return []
                    items = client.query(
                        TableName=collection_name,
                        ExpressionAttributeValues=query_param[
                            "ExpressionAttributeValues"
                        ],
                        KeyConditionExpression=query_param["KeyConditionExpression"],
                    )
                    for item in items.get("Items"):
                        result = {}
                        for key, value in item.items():
                            deserialized_value = deserializer.deserialize(value)
                            result[key] = deserialized_value
                        results.append(result)
            return results
        except ClientError as error:
            raise ConnectorFailureException(error.response["Error"]["Message"])

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """Execute a masking request for DynamoDB"""

        query_config = self.query_config(node)
        collection_name = node.address.collection
        update_ct = 0
        for row in rows:
            update_items = query_config.generate_update_stmt(
                row, policy, privacy_request
            )
            if update_items is not None:
                client = self.client()
                update_result = client.put_item(
                    TableName=collection_name,
                    Item=update_items,
                )
                if update_result["ResponseMetadata"]["HTTPStatusCode"] == 200:
                    update_ct += 1
                logger.info(
                    "client.put_item({}, {})",
                    collection_name,
                    Pii(update_items),
                )

        return update_ct


def product_dict(**kwargs: List) -> Generator:
    """
    Takes a dictionary of lists, returning the product
    as a list of dictionaries.
    """
    keys = kwargs.keys()
    for instance in itertools.product(*kwargs.values()):
        yield dict(zip(keys, instance))



"""This module handles the logic for syncing remote resource versions into their local file."""

from typing import Any, Dict, List, Optional

import yaml
from fideslang import model_list
from fideslang.manifests import load_yaml_into_dict

from fides.common.utils import echo_green, echo_red, print_divider
from fides.core.api_helpers import get_server_resource, list_server_resources
from fides.core.utils import get_manifest_list

MODEL_LIST = model_list


def remove_nulls(obj: Any) -> Any:
    """
    Recursively remove all null values from dictionaries and lists.

    Args:
        obj: Any Python object (dict, list, etc.)

    Returns:
        The transformed data structure with null values removed
    """
    if isinstance(obj, list):
        # For lists: process each item and filter out None values
        return [remove_nulls(item) for item in obj if item is not None]

    if isinstance(obj, dict):
        # For dictionaries: process each key-value pair and filter out pairs with None values
        return {
            key: remove_nulls(value) for key, value in obj.items() if value is not None
        }

    # Return all other data types unchanged
    return obj


def write_manifest_file(manifest_path: str, manifest: Dict) -> None:
    """
    Write a manifest file out.
    """
    with open(manifest_path, "w", encoding="utf-8") as manifest_file:
        yaml.dump(manifest, manifest_file, sort_keys=False, indent=2)
    echo_green(f"Updated manifest file written out to: '{manifest_path}'")


def pull_existing_resources(
    manifests_dir: str,
    url: str,
    headers: Dict[str, str],
) -> List[str]:
    """
    Update all of the pre-existing local resources to match their
    state on the server.
    """
    manifest_path_list = get_manifest_list(manifests_dir)
    # Store and return the keys of resources that get pulled here.
    existing_keys: List[str] = []

    print_divider()
    for manifest_path in manifest_path_list:
        print(f"Pulling file: '{manifest_path}'...")
        manifest = load_yaml_into_dict(manifest_path)
        updated_manifest = {}

        for resource_type in manifest.keys():
            resource_list = manifest[resource_type]
            updated_resource_list = []

            for resource in resource_list:
                fides_key = resource["fides_key"]
                existing_keys.append(fides_key)

                server_resource = get_server_resource(
                    url, resource_type, fides_key, headers
                )

                if server_resource:
                    # Remove null values from the server resource
                    server_resource = remove_nulls(server_resource)
                    updated_resource_list.append(server_resource)
                    print(
                        f" - {resource_type.capitalize()} with fides_key: {fides_key} is being updated from the server..."
                    )
                else:
                    updated_resource_list.append(resource)

            updated_manifest[resource_type] = updated_resource_list
            write_manifest_file(manifest_path, updated_manifest)
        print_divider()

    return existing_keys


def pull_resource_by_key(
    manifests_dir: str,
    url: str,
    headers: Dict[str, str],
    fides_key: str,
    resource_type: str,
) -> None:
    """
    Pull a resource from the server by its fides_key and update the local manifest file if it exists,
    otherwise a new manifest file at {manifests_dir}/{resource_type}.yml
    """
    if manifests_dir[-1] == "/":
        manifests_dir = manifests_dir[:-1]
    manifest_path = f"{manifests_dir}/{fides_key}.yml"
    print(f"Pulling {resource_type} with fides_key: {fides_key}...", end="\n")
    server_resource = get_server_resource(url, resource_type, fides_key, headers)

    if server_resource:
        # Remove null values from the server resource
        server_resource = remove_nulls(server_resource)
        try:
            manifest = load_yaml_into_dict(manifest_path)
        except FileNotFoundError:
            print(
                f"Manifest file {manifest_path} does not already exist and will be created"
            )
            manifest = {}
        manifest[resource_type] = [server_resource]
        write_manifest_file(manifest_path, manifest)

    else:
        echo_red(
            f"{resource_type} with fides_key: {fides_key} not found on the server."
        )


def pull_missing_resources(
    manifest_path: str,
    url: str,
    headers: Dict[str, str],
    existing_keys: List[str],
) -> bool:
    """
    Writes all "system", "dataset" and "policy" resources out locally
    that currently only exist on the server.
    """

    print(f"Writing out new resources to file: '{manifest_path}'...")
    resource_manifest = {
        resource: list_server_resources(
            url=url,
            headers=headers,
            resource_type=resource,
            exclude_keys=existing_keys,
        )
        for resource in MODEL_LIST
    }

    # Remove null values from all resources
    for resource_type, resources in resource_manifest.items():
        resource_manifest[resource_type] = [
            remove_nulls(resource) for resource in (resources or [])
        ]

    resource_manifest = {
        key: value for key, value in resource_manifest.items() if value
    }

    # Write out the resources in a file
    write_manifest_file(manifest_path, resource_manifest)
    print_divider()
    return True


def pull(
    manifests_dir: str,
    url: str,
    headers: Dict[str, str],
    fides_key: Optional[str],
    resource_type: Optional[str],
    all_resources_file: Optional[str],
) -> None:
    """
    If a resource in a local file has a matching resource on the server,
    write the server version into the local file.

    If the 'all_resources_file' option is present, pull all other server resources
    into a local file.

    If only `fides_key` is provided, only that resource will be pulled.
    """

    if fides_key and resource_type:
        pull_resource_by_key(
            manifests_dir=manifests_dir,
            url=url,
            headers=headers,
            fides_key=fides_key,
            resource_type=resource_type,
        )
        echo_green("Pull complete.")
        return

    existing_keys = pull_existing_resources(
        manifests_dir=manifests_dir,
        url=url,
        headers=headers,
    )

    if all_resources_file:
        pull_missing_resources(
            manifest_path=all_resources_file,
            url=url,
            headers=headers,
            existing_keys=existing_keys,
        )

    echo_green("Pull complete.")



# pylint: disable=R0401, C0302

from __future__ import annotations

import json
from datetime import datetime
from enum import Enum as EnumType
from typing import Optional

from pydantic import BaseModel, ConfigDict

from fides.api.models.policy import PolicyPreWebhook, WebhookDirection
from fides.api.models.pre_approval_webhook import PreApprovalWebhook
from fides.api.models.privacy_request.request_task import RequestTask
from fides.api.oauth.jwt import generate_jwe
from fides.api.schemas.external_https import (
    DownloadTokenJWE,
    RequestTaskJWE,
    WebhookJWE,
)
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.schemas.redis_cache import Identity
from fides.common.api.scope_registry import (
    PRIVACY_REQUEST_CALLBACK_RESUME,
    PRIVACY_REQUEST_READ_ACCESS_RESULTS,
    PRIVACY_REQUEST_REVIEW,
)
from fides.config import CONFIG


class CallbackType(EnumType):
    """We currently have three types of Webhooks: pre-approval, pre (-execution), post (-execution)"""

    pre_approval = "pre_approval"
    pre = "pre"  # pre-execution
    post = "post"  # post-execution


class SecondPartyRequestFormat(BaseModel):
    """
    The request body we will use when calling a user's HTTP endpoint from fides.api
    This class is defined here to avoid circular import issues between this file and
    models.policy
    """

    privacy_request_id: str
    privacy_request_status: PrivacyRequestStatus
    direction: WebhookDirection
    callback_type: CallbackType
    identity: Identity
    policy_action: Optional[ActionType] = None
    model_config = ConfigDict(use_enum_values=True)


def generate_request_callback_resume_jwe(webhook: PolicyPreWebhook) -> str:
    """
    Generate a JWE to be used to resume privacy request execution.
    """
    jwe = WebhookJWE(
        webhook_id=webhook.id,
        scopes=[PRIVACY_REQUEST_CALLBACK_RESUME],
        iat=datetime.now().isoformat(),
    )
    return generate_jwe(
        json.dumps(jwe.model_dump(mode="json")),
        CONFIG.security.app_encryption_key,
    )


def generate_request_callback_pre_approval_jwe(webhook: PreApprovalWebhook) -> str:
    """
    Generate a JWE to be used to mark privacy requests as eligible / not-eligible for pre approval.
    """
    jwe = WebhookJWE(
        webhook_id=webhook.id,
        scopes=[PRIVACY_REQUEST_REVIEW],
        iat=datetime.now().isoformat(),
    )
    return generate_jwe(
        json.dumps(jwe.model_dump(mode="json")),
        CONFIG.security.app_encryption_key,
    )


def generate_request_task_callback_jwe(request_task: RequestTask) -> str:
    """
    Generate a JWE to be used to resume privacy request execution when a
    callback endpoint is hit for a RequestTask
    """
    jwe = RequestTaskJWE(
        request_task_id=request_task.id,
        scopes=[PRIVACY_REQUEST_CALLBACK_RESUME],
        iat=datetime.now().isoformat(),
    )
    return generate_jwe(
        json.dumps(jwe.model_dump(mode="json")),
        CONFIG.security.app_encryption_key,
    )


def generate_privacy_request_download_token(privacy_request_id: str) -> str:
    """
    Generate a JWE token for users to download their privacy request access package.
    This is currently used for the DSR package link endpoint which provides a redirect
    to a presigned URL for the access results.
    This token expires based on the configured TTL for security.
    """
    from datetime import timedelta

    now = datetime.now()
    # Use the configured TTL from security settings
    expiration = now + timedelta(
        seconds=CONFIG.security.subject_request_download_link_ttl_seconds
    )

    jwe = DownloadTokenJWE(
        privacy_request_id=privacy_request_id,
        scopes=[PRIVACY_REQUEST_READ_ACCESS_RESULTS],
        iat=now.isoformat(),
        exp=expiration.isoformat(),
    )
    return generate_jwe(
        json.dumps(jwe.model_dump(mode="json")),
        CONFIG.security.app_encryption_key,
    )



from enum import Enum
from typing import TYPE_CHECKING, Optional, Union

from sqlalchemy import Column, ForeignKey, Index, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base, FidesBase
from fides.api.db.util import EnumColumn
from fides.api.task.conditional_dependencies.schemas import (
    ConditionGroup,
    ConditionLeaf,
)

if TYPE_CHECKING:
    from fides.api.models.manual_task.manual_task import ManualTask


class ManualTaskConditionalDependencyType(str, Enum):
    """Enum for manual task conditional dependency types.

    This enum defines the two types of nodes in a conditional dependency tree:

    - leaf: A terminal node that represents a single condition (e.g., "user.age >= 18")
    - group: A non-terminal node that groups multiple conditions with logical operators (AND/OR)

    Examples:
        leaf: Used for simple field comparisons like:
            - "user.name exists"
            - "user.age >= 18"
            - "billing.subscription.status == 'active'"

        group: Used to combine multiple conditions with logical operators:
            - AND group: "user.age >= 18 AND user.active == true"
            - OR group: "user.role == 'admin' OR user.verified == true"
            - Nested groups: "(user.age >= 18 AND (user.role == 'admin' OR user.verified == true))"
    """

    leaf = "leaf"
    group = "group"


class ManualTaskConditionalDependency(Base):
    """Model for storing conditional dependencies."""

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "manual_task_conditional_dependency"

    # We need to redefine it here so that self-referential relationships
    # can properly reference the `id` column instead of the built-in Python function.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)

    # Foreign key relationships
    manual_task_id = Column(
        String, ForeignKey("manual_task.id", ondelete="CASCADE"), nullable=False
    )
    parent_id = Column(
        String,
        ForeignKey("manual_task_conditional_dependency.id", ondelete="CASCADE"),
        nullable=True,
    )

    # Condition metadata
    condition_type = Column(
        EnumColumn(ManualTaskConditionalDependencyType), nullable=False
    )  # leaf or group
    field_address = Column(String, nullable=True)  # For leaf conditions
    operator = Column(String, nullable=True)  # For leaf conditions
    value = Column(JSONB, nullable=True)  # For leaf conditions
    logical_operator = Column(String, nullable=True)  # 'and' or 'or' for groups

    # Ordering
    sort_order = Column(Integer, nullable=False, default=0)

    __table_args__ = (
        Index("ix_manual_task_conditional_dependency_manual_task_id", "manual_task_id"),
        Index("ix_manual_task_conditional_dependency_parent_id", "parent_id"),
        Index("ix_manual_task_conditional_dependency_condition_type", "condition_type"),
        Index("ix_manual_task_conditional_dependency_sort_order", "sort_order"),
    )

    # Relationships
    task = relationship("ManualTask", back_populates="conditional_dependencies")
    parent = relationship(
        "ManualTaskConditionalDependency",
        remote_side=[id],
        back_populates="children",
        foreign_keys=[parent_id],
    )
    children = relationship(
        "ManualTaskConditionalDependency",
        back_populates="parent",
        cascade="all, delete-orphan",
        foreign_keys=[parent_id],
    )

    def to_condition_leaf(self) -> ConditionLeaf:
        """Convert to ConditionLeaf if this is a leaf condition"""
        if self.condition_type != "leaf":
            raise ValueError("Cannot convert group condition to leaf")

        return ConditionLeaf(
            field_address=self.field_address, operator=self.operator, value=self.value
        )

    def to_condition_group(self) -> ConditionGroup:
        """Convert to ConditionGroup if this is a group condition"""
        if self.condition_type != "group":
            raise ValueError("Cannot convert leaf condition to group")

        # Recursively build children
        child_conditions = []
        children_list = [child for child in self.children]  # type: ignore[attr-defined]
        for child in sorted(children_list, key=lambda x: x.sort_order):
            if child.condition_type == "leaf":
                child_conditions.append(child.to_condition_leaf())
            else:
                child_conditions.append(child.to_condition_group())

        return ConditionGroup(
            logical_operator=self.logical_operator, conditions=child_conditions
        )

    @classmethod
    def get_root_condition(
        cls, db: Session, manual_task_id: str
    ) -> Optional[Union[ConditionLeaf, ConditionGroup]]:
        """Get the root condition for a config"""
        root = (
            db.query(cls)
            .filter(cls.manual_task_id == manual_task_id, cls.parent_id.is_(None))
            .first()
        )

        if not root:
            return None

        if root.condition_type == "leaf":
            return root.to_condition_leaf()

        return root.to_condition_group()



from requests import PreparedRequest

from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.schemas.saas.strategy_configuration import OAuth2BaseConfiguration
from fides.api.service.authentication.authentication_strategy_oauth2_base import (
    OAuth2AuthenticationStrategyBase,
)


class OAuth2ClientCredentialsAuthenticationStrategy(OAuth2AuthenticationStrategyBase):
    """
    Checks the expiration date on the stored access token and refreshes
    it if needed using the configured token refresh request.
    """

    name = "oauth2_client_credentials"
    configuration_model = OAuth2BaseConfiguration

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """
        Checks the expiration date on the existing access token and refreshes if necessary.
        The existing/updated access token is then added to the request as a bearer token.
        """

        access_token = connection_config.secrets.get("access_token")  # type: ignore
        if not access_token:
            access_token = self.get_access_token(connection_config)
        else:
            access_token = self._refresh_token(connection_config)

        # add access_token to request
        request.headers["Authorization"] = "Bearer " + access_token
        return request

    def _refresh_token(self, connection_config: ConnectionConfig) -> str:
        """
        Persists and returns a refreshed access_token if the token is close to expiring.
        Otherwise just returns the existing access_token.

        For the Client Credentials OAuth flow we reuse the access request to get a new token.
        """

        expires_at = connection_config.secrets.get("expires_at")  # type: ignore
        if self._close_to_expiration(expires_at, connection_config):
            refresh_response = self._call_token_request(
                "refresh", self.token_request, connection_config
            )
            return self._validate_and_store_response(
                refresh_response, connection_config
            )

        return connection_config.secrets.get("access_token")  # type: ignore



"""add okta to connection_types

Revision ID: d0cbfec0b2dd
Revises: c9c72b3d550b
Create Date: 2025-04-23 15:22:57.920715

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d0cbfec0b2dd"
down_revision = "c9c72b3d550b"
branch_labels = None
depends_on = None

type_to_handle = "okta"


def get_enum_values():
    conn = op.get_bind()
    result = conn.execute("SELECT unnest(enum_range(NULL::connectiontype))").fetchall()
    return [row[0] for row in result]


def upgrade():
    # add 'okta' to ConnectionType enum
    enum_values = [f"'{value}'" for value in get_enum_values() + [type_to_handle]]
    enum_values.sort()  # Just to keep it fantastic

    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_old")
    op.execute(f"CREATE TYPE connectiontype AS ENUM ({', '.join(set(enum_values))})")
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_old")


def downgrade():
    # Remove 'okta' from ConnectionType enum
    enum_values = [f"'{v.strip()}'" for v in get_enum_values() if v != type_to_handle]
    enum_values.sort()  # Just to keep it fantastic

    op.execute(
        f"DELETE FROM connectionconfig WHERE connection_type IN ('{type_to_handle}')"
    )
    op.execute("ALTER TYPE connectiontype RENAME TO connectiontype_old")

    op.execute(f"CREATE TYPE connectiontype AS ENUM ({', '.join(set(enum_values))})")
    op.execute(
        """
        ALTER TABLE connectionconfig ALTER COLUMN connection_type TYPE connectiontype USING
        connection_type::text::connectiontype
    """
    )
    op.execute("DROP TYPE connectiontype_old")



from typing import Set

from pygtrie import StringTrie


def find_undeclared_categories(
    input_categories: Set[str], declared_categories: Set[str]
) -> Set[str]:
    """
    Finds the categories in input_categories that are not declared in the given list of declared categories.

    This function uses a trie-based approach to perform hierarchical matching of categories. It considers
    a category as declared if it matches any of the following conditions:
    - The category itself is present in the declared categories.
    - Any parent category of the category is present in the declared categories.

    For example, if "user.contact.email" is in input_categories and the declared categories contain
    "user.contact" or "user", then "user.contact.email" will be considered as declared.
    """

    declared_trie = StringTrie(separator=".")
    for category in declared_categories:
        declared_trie[category] = True

    undeclared_categories = set()
    for category in input_categories:
        prefix, _ = declared_trie.longest_prefix(category)
        if not prefix:
            undeclared_categories.add(category)

    return undeclared_categories



from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict

from fides.api.db.base_class import Base


class FidesCloud(Base):
    """Stores all Fides Cloud related config variables"""

    config = Column(MutableDict.as_mutable(JSONB), nullable=True)



from typing import List

import pymysql
from google.cloud.sql.connector import Connector
from google.oauth2 import service_account
from sqlalchemy.engine import Engine, LegacyCursorResult, create_engine  # type: ignore

from fides.api.schemas.connection_configuration.connection_secrets_google_cloud_sql_mysql import (
    GoogleCloudSQLMySQLSchema,
)
from fides.api.schemas.connection_configuration.enums.google_cloud_sql_ip_type import (
    GoogleCloudSQLIPType,
)
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row
from fides.config import get_config

CONFIG = get_config()


class GoogleCloudSQLMySQLConnector(SQLConnector):
    """Connector specific to Google Cloud SQL for MySQL"""

    secrets_schema = GoogleCloudSQLMySQLSchema

    # Overrides SQLConnector.create_client
    def create_client(self) -> Engine:
        """Returns a SQLAlchemy Engine that can be used to interact with a database"""

        config = self.secrets_schema(**self.configuration.secrets or {})

        credentials = service_account.Credentials.from_service_account_info(
            dict(config.keyfile_creds)
        )

        # initialize connector with the loaded credentials
        connector = Connector(credentials=credentials)

        def getconn() -> pymysql.connections.Connection:
            conn: pymysql.connections.Connection = connector.connect(
                config.instance_connection_name,
                "pymysql",
                ip_type=config.ip_type or GoogleCloudSQLIPType.public,
                user=config.db_iam_user,
                db=config.dbname,
                enable_iam_auth=True,
            )
            return conn

        return create_engine("mysql+pymysql://", creator=getconn)

    @staticmethod
    def cursor_result_to_rows(results: LegacyCursorResult) -> List[Row]:
        """results to a list of dictionaries"""
        return SQLConnector.default_cursor_result_to_rows(results)

    def build_uri(self) -> None:
        """
        We need to override this method so it is not abstract anymore, and GoogleCloudSQLMySQLConnector is instantiable.
        """



from typing import Any, Dict, List, Optional

import jwt
from fastapi import Depends, HTTPException, Security
from loguru import logger
from sqlalchemy.orm import Session
from starlette.status import (
    HTTP_200_OK,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_424_FAILED_DEPENDENCY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from fides.api import common_exceptions
from fides.api.api import deps
from fides.api.api.v1.endpoints.privacy_request_endpoints import (
    get_privacy_request_or_error,
)
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.drp_privacy_request import (
    DRP_VERSION,
    DrpDataRightsResponse,
    DrpIdentity,
    DrpPrivacyRequestCreate,
    DrpRevokeRequest,
)
from fides.api.schemas.policy import DrpAction
from fides.api.schemas.privacy_request import (
    PrivacyRequestDRPStatusResponse,
    PrivacyRequestStatus,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.service.drp.drp_fidesops_mapper import DrpFidesopsMapper
from fides.api.service.privacy_request.request_service import (
    build_required_privacy_request_kwargs,
    cache_data,
)
from fides.api.util.api_router import APIRouter
from fides.api.util.cache import FidesopsRedis
from fides.api.util.logger import Pii
from fides.common.api import scope_registry as scopes
from fides.common.api.v1 import urn_registry as urls
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy
from fides.service.messaging.messaging_service import (
    check_and_dispatch_error_notifications,
)
from fides.service.privacy_request.privacy_request_service import queue_privacy_request

router = APIRouter(tags=["DRP"], prefix=urls.V1_URL_PREFIX)


EMBEDDED_EXECUTION_LOG_LIMIT = 50


@router.post(
    urls.DRP_EXERCISE,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestDRPStatusResponse,
)
async def create_drp_privacy_request(
    *,
    cache: FidesopsRedis = Depends(deps.get_cache),
    db: Session = Depends(deps.get_db),
    config_proxy: ConfigProxy = Depends(deps.get_config_proxy),
    data: DrpPrivacyRequestCreate,
) -> PrivacyRequestDRPStatusResponse:
    """
    Given a drp privacy request body, create and execute
    a corresponding Fidesops PrivacyRequest
    """

    jwt_key = CONFIG.security.drp_jwt_secret
    if jwt_key is None:
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail="JWT key must be provided",
        )

    logger.info("Finding policy with drp action '{}'", data.exercise[0])
    policy: Optional[Policy] = Policy.get_by(
        db=db,
        field="drp_action",
        value=data.exercise[0],  # type: ignore[arg-type]
    )

    if not policy:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No policy found with drp action '{data.exercise}'.",
        )

    privacy_request_kwargs: Dict[str, Any] = build_required_privacy_request_kwargs(
        None,
        policy.id,
        config_proxy.execution.subject_identity_verification_required,
        authenticated=False,
    )

    try:
        decrypted_identity: DrpIdentity = DrpIdentity(
            **jwt.decode(data.identity, jwt_key, algorithms=["HS256"])
        )

        mapped_identity: Identity = DrpFidesopsMapper.map_identity(
            drp_identity=decrypted_identity
        )

        privacy_request: PrivacyRequest = PrivacyRequest.create(
            db=db,
            data=privacy_request_kwargs,
        )
        privacy_request.persist_identity(
            db=db,
            identity=mapped_identity,
        )

        check_and_dispatch_error_notifications(db=db)

        logger.info(
            "Decrypting identity for DRP privacy request {}", privacy_request.id
        )

        cache_data(privacy_request, mapped_identity, None, data)

        if masking_secrets := policy.generate_masking_secrets():
            logger.info(
                "Caching masking secrets for privacy request {}", privacy_request.id
            )
            privacy_request.persist_masking_secrets(masking_secrets)

        queue_privacy_request(privacy_request.id)

        return PrivacyRequestDRPStatusResponse(
            request_id=privacy_request.id,
            received_at=privacy_request.requested_at,
            status=DrpFidesopsMapper.map_status(privacy_request.status),  # type: ignore
        )

    except common_exceptions.RedisConnectionError as exc:
        logger.error("RedisConnectionError: {}", Pii(str(exc)))
        # Thrown when cache.ping() fails on cache connection retrieval
        raise HTTPException(
            status_code=HTTP_424_FAILED_DEPENDENCY,
            detail=exc.args[0],
        )
    except Exception as exc:
        logger.error("Exception: {}", Pii(str(exc)))
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="DRP privacy request could not be exercised",
        )


@router.get(
    urls.DRP_STATUS,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.PRIVACY_REQUEST_READ])],
    response_model=PrivacyRequestDRPStatusResponse,
)
def get_request_status_drp(
    *, db: Session = Depends(deps.get_db), request_id: str
) -> PrivacyRequestDRPStatusResponse:
    """
    Returns PrivacyRequest information where the respective privacy request is associated with
    a policy that implements a Data Rights Protocol action.
    """

    logger.info("Finding request for DRP with ID: {}", request_id)
    request = PrivacyRequest.get(
        db=db,
        object_id=request_id,
    )
    if not request or not request.policy or not request.policy.drp_action:
        # If no request is found with this ID, or that request has no policy,
        # or that request's policy has no associated drp_action.
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"Privacy request with ID {request_id} does not exist, or is not associated with a data rights protocol action.",
        )

    logger.info("Privacy request with ID: {} found for DRP status.", request_id)
    return PrivacyRequestDRPStatusResponse(
        request_id=request.id,
        received_at=request.requested_at,
        status=DrpFidesopsMapper.map_status(request.status),
    )


@router.get(
    urls.DRP_DATA_RIGHTS,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.POLICY_READ])],
    response_model=DrpDataRightsResponse,
)
def get_drp_data_rights(*, db: Session = Depends(deps.get_db)) -> DrpDataRightsResponse:
    """
    Query all policies and determine the list of DRP actions that are attached to existing policies.
    """

    logger.info("Fetching available DRP data rights")
    actions: List[DrpAction] = [
        item.drp_action  # type: ignore
        for item in db.query(Policy.drp_action).filter(Policy.drp_action.isnot(None))
    ]

    return DrpDataRightsResponse(
        version=DRP_VERSION, api_base=None, actions=actions, user_relationships=None
    )


@router.post(
    urls.DRP_REVOKE,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.PRIVACY_REQUEST_REVIEW])
    ],
    response_model=PrivacyRequestDRPStatusResponse,
)
def revoke_request(
    *, db: Session = Depends(deps.get_db), data: DrpRevokeRequest
) -> PrivacyRequestDRPStatusResponse:
    """
    Revoke a pending privacy request.
    """
    privacy_request: PrivacyRequest = get_privacy_request_or_error(db, data.request_id)

    if privacy_request.status != PrivacyRequestStatus.pending:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Invalid revoke request. Can only revoke `pending` requests. Privacy request '{privacy_request.id}' status = {privacy_request.status.value}.",  # type: ignore
        )

    logger.info("Canceling privacy request '{}'", privacy_request.id)
    privacy_request.cancel_processing(db, cancel_reason=data.reason)

    return PrivacyRequestDRPStatusResponse(
        request_id=privacy_request.id,
        received_at=privacy_request.requested_at,
        status=DrpFidesopsMapper.map_status(privacy_request.status),  # type: ignore
        reason=data.reason,
    )



"""manual task allow deletion with orphaned configs

Revision ID: 41a634d8c669
Revises: aadfe83c5644
Create Date: 2025-06-22 19:34:22.336406

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "41a634d8c669"
down_revision = "aadfe83c5644"
branch_labels = None
depends_on = None


def upgrade():
    """Allow manual task deletion by making task_id nullable and using SET NULL constraints."""

    # Make task_id columns nullable to support orphaned records
    op.alter_column("manual_task_config", "task_id", nullable=True)
    op.alter_column("manual_task_instance", "task_id", nullable=True)
    op.alter_column("manual_task_submission", "task_id", nullable=True)
    op.alter_column("manual_task_config_field", "task_id", nullable=True)
    op.alter_column("manual_task_log", "task_id", nullable=True)

    # manual_task_config: Change task_id to SET NULL (preserve configs when task deleted)
    op.drop_constraint(
        "manual_task_config_task_id_fkey", "manual_task_config", type_="foreignkey"
    )
    op.create_foreign_key(
        None,
        "manual_task_config",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="SET NULL",
    )

    # manual_task_instance: Change task_id to SET NULL (preserve instances when task deleted)
    # Keep config_id as RESTRICT (instances should not be deleted when config deleted)
    op.drop_constraint(
        "manual_task_instance_config_id_fkey",
        "manual_task_instance",
        type_="foreignkey",
    )
    op.drop_constraint(
        "manual_task_instance_task_id_fkey", "manual_task_instance", type_="foreignkey"
    )
    op.create_foreign_key(
        None,
        "manual_task_instance",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        None,
        "manual_task_instance",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="SET NULL",
    )

    # manual_task_submission: Change task_id to SET NULL, keep others as RESTRICT
    # Keep instance_id as CASCADE since submissions should be deleted when instance is deleted
    op.drop_constraint(
        "manual_task_submission_config_id_fkey",
        "manual_task_submission",
        type_="foreignkey",
    )
    op.drop_constraint(
        "manual_task_submission_field_id_fkey",
        "manual_task_submission",
        type_="foreignkey",
    )
    op.drop_constraint(
        "manual_task_submission_task_id_fkey",
        "manual_task_submission",
        type_="foreignkey",
    )
    op.create_foreign_key(
        None,
        "manual_task_submission",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.create_foreign_key(
        None,
        "manual_task_submission",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        None,
        "manual_task_submission",
        "manual_task_config_field",
        ["field_id"],
        ["id"],
        ondelete="RESTRICT",
    )

    # manual_task_config_field: Change task_id to SET NULL (preserve fields when task deleted)
    op.drop_constraint(
        "manual_task_config_field_task_id_fkey",
        "manual_task_config_field",
        type_="foreignkey",
    )
    op.create_foreign_key(
        None,
        "manual_task_config_field",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="SET NULL",
    )

    # manual_task_log: Change task_id to SET NULL (preserve logs when task deleted)
    op.drop_constraint(
        "manual_task_log_task_id_fkey", "manual_task_log", type_="foreignkey"
    )
    op.create_foreign_key(
        "manual_task_log_task_id_fkey",
        "manual_task_log",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    """Revert to original CASCADE behavior and make task_id non-nullable."""

    # Revert manual_task_log constraints
    op.drop_constraint(
        "manual_task_log_task_id_fkey", "manual_task_log", type_="foreignkey"
    )
    op.create_foreign_key(
        "manual_task_log_task_id_fkey",
        "manual_task_log",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="CASCADE",
    )

    # Revert manual_task_config_field constraints
    op.drop_constraint(
        "manual_task_config_field_task_id_fkey",
        "manual_task_config_field",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "manual_task_config_field_task_id_fkey",
        "manual_task_config_field",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="CASCADE",
    )

    # Revert manual_task_submission constraints
    op.drop_constraint(
        "manual_task_submission_task_id_fkey",
        "manual_task_submission",
        type_="foreignkey",
    )
    op.drop_constraint(
        "manual_task_submission_field_id_fkey",
        "manual_task_submission",
        type_="foreignkey",
    )
    op.drop_constraint(
        "manual_task_submission_config_id_fkey",
        "manual_task_submission",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "manual_task_submission_task_id_fkey",
        "manual_task_submission",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "manual_task_submission_field_id_fkey",
        "manual_task_submission",
        "manual_task_config_field",
        ["field_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "manual_task_submission_config_id_fkey",
        "manual_task_submission",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="CASCADE",
    )

    # Revert manual_task_instance constraints
    op.drop_constraint(
        "manual_task_instance_task_id_fkey", "manual_task_instance", type_="foreignkey"
    )
    op.drop_constraint(
        "manual_task_instance_config_id_fkey",
        "manual_task_instance",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "manual_task_instance_task_id_fkey",
        "manual_task_instance",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "manual_task_instance_config_id_fkey",
        "manual_task_instance",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="CASCADE",
    )

    # Revert manual_task_config constraints
    op.drop_constraint(
        "manual_task_config_task_id_fkey", "manual_task_config", type_="foreignkey"
    )
    op.create_foreign_key(
        "manual_task_config_task_id_fkey",
        "manual_task_config",
        "manual_task",
        ["task_id"],
        ["id"],
        ondelete="CASCADE",
    )

    # Make task_id columns non-nullable again
    op.alter_column("manual_task_log", "task_id", nullable=False)
    op.alter_column("manual_task_config_field", "task_id", nullable=False)
    op.alter_column("manual_task_submission", "task_id", nullable=False)
    op.alter_column("manual_task_instance", "task_id", nullable=False)
    op.alter_column("manual_task_config", "task_id", nullable=False)



# pylint: disable=protected-access
from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Type

from sqlalchemy import (
    BOOLEAN,
    Column,
    ForeignKey,
    ForeignKeyConstraint,
    String,
    Text,
    UniqueConstraint,
)
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import RelationshipProperty, Session, relationship

from fides.api.common_exceptions import ValidationError
from fides.api.db.base_class import Base
from fides.api.models.sql_models import FidesBase  # type: ignore[attr-defined]

LEGACY_TAXONOMIES = {"data_categories", "data_uses", "data_subjects"}


class TargetType(str, Enum):
    """Enumeration of target types that taxonomies can apply to."""

    SYSTEM = "system"
    PRIVACY_DECLARATION = "privacy_declaration"
    TAXONOMY = "taxonomy"  # For taxonomy-to-taxonomy relationships


class Taxonomy(Base, FidesBase):
    """The SQL model for taxonomy resources.

    This is a generic taxonomy model that can be used to create any taxonomy.
    For now we seed the database with the legacy taxonomies (data_category, data use, data subject)
    so that these legacy taxonomy types can be used for allowed usage relationships.
    """

    # Overriding the id definition from Base so we don't treat this as the primary key
    id = Column(
        String(255),
        nullable=False,
        index=False,
        unique=True,
        default=FidesBase.generate_uuid,
    )

    # The fides_key is inherited from FidesBase and acts as the primary key

    # This is private to encourage the use of applies_to (see comment below)
    _allowed_usages: RelationshipProperty[List[TaxonomyAllowedUsage]] = relationship(
        "TaxonomyAllowedUsage",
        back_populates="source_taxonomy",
        cascade="all, delete-orphan",
        lazy="selectin",
    )

    # Association proxy to simplify access to target_type values
    # This allows getting a list of strings while the actual storage
    # is handled through the TaxonomyAllowedUsage model
    # Updates should be done through the create/update methods
    applies_to: List[str] = association_proxy(
        "_allowed_usages",
        "target_type",
        # Allow setting via strings, the relationship backref will set FK
        creator=lambda target_type: TaxonomyAllowedUsage(target_type=target_type),
    )

    @classmethod
    def create(
        cls: Type["Taxonomy"],
        db: Session,
        *,
        data: Dict[str, Any],
        check_name: bool = True,
    ) -> "Taxonomy":
        """Create a new Taxonomy with proper handling of applies_to."""
        # Disallow creating taxonomies that represent legacy types
        fides_key = data.get("fides_key")
        if fides_key in LEGACY_TAXONOMIES:
            raise ValidationError(
                f"Cannot create taxonomy '{fides_key}'. This is a taxonomy managed by the system."
            )
        applies_to = data.get("applies_to", [])

        # Create a copy of data without applies_to to avoid modifying the input
        data_without_applies_to = {
            key: value for key, value in data.items() if key != "applies_to"
        }

        # Create the taxonomy
        taxonomy: Taxonomy = super().create(
            db=db, data=data_without_applies_to, check_name=check_name
        )

        # Reconcile allowed usages if applies_to was provided
        if applies_to:
            taxonomy._reconcile_allowed_usages(db, applies_to)

        return cls.persist_obj(db, taxonomy)

    def update(self, db: Session, *, data: Dict[str, Any]) -> "Taxonomy":
        """Update a Taxonomy with proper handling of applies_to."""
        applies_to = data.get("applies_to", None)

        # Create a copy of data without applies_to to avoid modifying the input
        data_without_applies_to = {
            key: value for key, value in data.items() if key != "applies_to"
        }

        # Update the base fields
        super().update(db=db, data=data_without_applies_to)

        # If applies_to was provided, reconcile allowed usages
        if applies_to is not None:
            self._reconcile_allowed_usages(db, applies_to)

        return self

    def save(self, db: Session) -> "Taxonomy":
        """Override save to reconcile any direct `applies_to` edits before persisting.

        This allows callers to mutate `applies_to` via the association proxy and then call save.
        """
        # Ensure no duplicate target types and reconcile relationship objects to the current values
        self._reconcile_allowed_usages(db, list(self.applies_to))
        return super().save(db)  # type: ignore[return-value]

    def _reconcile_allowed_usages(self, db: Session, applies_to: List[str]) -> None:
        """Ensure `_allowed_usages` matches the provided `applies_to` list.

        - Deletes usages not in the provided list
        - Creates usages missing from the relationship
        - Deduplicates by target_type
        """
        existing_usages = {usage.target_type: usage for usage in self._allowed_usages}
        desired_types = set(applies_to)

        # Delete usages that should no longer exist
        for target_type in set(existing_usages.keys()) - desired_types:
            # Remove from relationship; delete-orphan cascade will handle DB delete
            self._allowed_usages.remove(existing_usages[target_type])

        # Add missing usages
        for target_type in desired_types - set(existing_usages.keys()):
            self._allowed_usages.append(TaxonomyAllowedUsage(target_type=target_type))

        # Deduplicate any accidental duplicates in-memory
        seen: set[str] = set()
        for usage in list(self._allowed_usages):
            if usage.target_type in seen:
                # Remove duplicate from relationship; delete-orphan will handle DB
                self._allowed_usages.remove(usage)
            else:
                seen.add(usage.target_type)


class TaxonomyAllowedUsage(Base):
    """
    The SQL model for taxonomy allowed usage.
    Defines what types of targets a taxonomy can be applied to.

    target_type can be either:
    - A generic type: "system", "privacy_declaration", "taxonomy"
    - A taxonomy key: "data_categories", "data_uses", etc.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "taxonomy_allowed_usage"

    id = Column(
        String(255),
        nullable=False,
        index=False,
        unique=True,
        default=FidesBase.generate_uuid,
    )

    source_taxonomy: RelationshipProperty[Taxonomy] = relationship(
        "Taxonomy",
        back_populates="_allowed_usages",
    )

    source_taxonomy_key: Column[str] = Column(
        String,
        ForeignKey("taxonomy.fides_key", ondelete="CASCADE"),
        primary_key=True,
    )
    target_type: Column[str] = Column(
        String, primary_key=True
    )  # Can be "system", "dataset", OR a taxonomy key like "data_categories"


class TaxonomyElement(Base, FidesBase):
    """
    The SQL model for taxonomy elements.

    This is a generic taxonomy element model that can be used to create any taxonomy element.

    As of now the legacy taxonomy elements still exist in their own tables (ctl_data_categories, ctl_data_uses, ctl_data_subjects),
    but we can migrate them to this model in the future if needed.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "taxonomy_element"

    id = Column(
        String(255),
        nullable=False,
        index=False,
        unique=True,
        default=FidesBase.generate_uuid,
    )

    # Which taxonomy this element belongs to
    taxonomy_type = Column(
        String,
        ForeignKey("taxonomy.fides_key", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )

    parent_key = Column(
        Text, ForeignKey("taxonomy_element.fides_key", ondelete="RESTRICT"), index=True
    )
    active = Column(BOOLEAN, default=True, nullable=False, index=True)

    children: RelationshipProperty[List[TaxonomyElement]] = relationship(
        "TaxonomyElement",
        back_populates="parent",
        cascade="save-update, merge, refresh-expire",  # intentionally do not cascade deletes
        passive_deletes="all",
    )

    parent: RelationshipProperty[Optional[TaxonomyElement]] = relationship(
        "TaxonomyElement",
        back_populates="children",
        remote_side="TaxonomyElement.fides_key",
    )


class TaxonomyUsage(Base):
    """
    The SQL model for taxonomy usage.
    Tracks the application of taxonomy elements to other taxonomy elements.

    Example: Applying a "high" tag (from sensitivity taxonomy) to "user.contact.email" (from data_categories taxonomy).
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "taxonomy_usage"

    # The taxonomy element being applied (e.g., risk)
    source_element_key = Column(String, nullable=False, index=True)

    # The taxonomy element it's being applied to (e.g., a data category)
    target_element_key = Column(String, nullable=False, index=True)

    # Denormalized taxonomy types for validation and performance
    source_taxonomy = Column(String, nullable=False, index=True)
    target_taxonomy = Column(String, nullable=False, index=True)

    __table_args__ = (
        # Validate that this type of usage is allowed
        ForeignKeyConstraint(
            ["source_taxonomy", "target_taxonomy"],
            [
                "taxonomy_allowed_usage.source_taxonomy_key",
                "taxonomy_allowed_usage.target_type",
            ],
            ondelete="RESTRICT",
            name="fk_taxonomy_usage_allowed",
        ),
        # Prevent duplicate applications
        UniqueConstraint(
            "source_element_key",
            "target_element_key",
            name="uq_taxonomy_usage",
        ),
    )



"""
Legacy taxonomy handler for DataCategory, DataUse, and DataSubject models.
"""

from typing import Dict, List, Optional, Type, Union

from sqlalchemy.orm import Query, Session

from fides.api.models.sql_models import (  # type:ignore[attr-defined]
    DataCategory,
    DataSubject,
    DataUse,
)

from .base import TaxonomyHandler


class LegacyTaxonomyHandler(TaxonomyHandler):
    """Handler for legacy taxonomy models (DataCategory, DataUse, DataSubject)."""

    def __init__(self, db: Session, taxonomy_type: str) -> None:
        super().__init__(db, taxonomy_type)
        self.models = {
            "data_categories": DataCategory,
            "data_uses": DataUse,
            "data_subjects": DataSubject,
        }
        if taxonomy_type not in self.models:
            raise ValueError(
                f"Taxonomy type '{taxonomy_type}' not supported. Supported types: {list(self.models.keys())}"
            )

    def get_elements(
        self, active_only: bool, parent_key: Optional[str]
    ) -> List[Union[DataCategory, DataUse, DataSubject]]:
        model_class = self.get_model()
        query = self._build_query(model_class, active_only, parent_key)
        return query.all()

    def get_element(
        self, fides_key: str
    ) -> Optional[Union[DataCategory, DataUse, DataSubject]]:
        model_class = self.get_model()
        element = (
            self.db.query(model_class)
            .filter(model_class.fides_key == fides_key)
            .first()
        )
        return element

    def create_element(
        self, element_data: Dict
    ) -> Union[DataCategory, DataUse, DataSubject]:
        model_class = self.get_model()
        element = model_class.create(self.db, data=element_data)
        return element

    def update_element(
        self, fides_key: str, element_data: Dict
    ) -> Optional[Union[DataCategory, DataUse, DataSubject]]:
        model_class = self.get_model()
        element = (
            self.db.query(model_class)
            .filter(model_class.fides_key == fides_key)
            .first()
        )

        if not element:
            return None

        return element.update(db=self.db, data=element_data)

    def delete_element(self, fides_key: str) -> None:
        model_class = self.get_model()
        element = (
            self.db.query(model_class)
            .filter(model_class.fides_key == fides_key)
            .first()
        )

        if element:
            self.db.delete(element)

    def get_model(self) -> Type[Union[DataCategory, DataUse, DataSubject]]:
        return self.models[self.taxonomy_type]

    def _build_query(
        self, model_class: Type, active_only: bool, parent_key: Optional[str]
    ) -> Query:
        query = self.db.query(model_class)
        if active_only:
            query = query.filter(model_class.active.is_(True))
        if parent_key and hasattr(model_class, "parent_key"):
            query = query.filter(model_class.parent_key == parent_key)
        return query



"""disable_inactive_notices

Revision ID: f17f92237383
Revises: 671358247251
Create Date: 2023-09-14 13:35:38.474968

"""

import uuid
from typing import Set

from alembic import op
from sqlalchemy import text
from sqlalchemy.engine import Connection, ResultProxy
from sqlalchemy.sql.elements import TextClause

# revision identifiers, used by Alembic.
from fides.api.models.sql_models import DataUse

revision = "f17f92237383"
down_revision = "671358247251"
branch_labels = None
depends_on = None


def get_expanded_system_data_uses(bind: Connection) -> Set:
    """Get data uses across all systems and those data uses' parents

    Same results as System.get_data_uses(System.all(db), include_parents=True), but
    we don't want to use the System Sqlalchemy model within a migration
    """
    system_data_uses: ResultProxy = bind.execute(
        text("SELECT data_use from privacydeclaration;")
    )

    expanded_data_uses = set()
    for row in system_data_uses:
        data_use: str = row["data_use"]
        # Expand system data uses so we also include their parents
        expanded_data_uses.update(DataUse.get_parent_uses_from_key(data_use))
    return expanded_data_uses


def disable_privacy_notice_and_bump_version(
    bind: Connection, privacy_notice_id: str, current_version: int
) -> None:
    """Disable the PrivacyNotice and bump its version"""
    disable_privacy_notice_query: TextClause = text(
        "UPDATE privacynotice SET disabled=true, version=:new_version WHERE id = :privacy_notice_id;"
    )
    bind.execute(
        disable_privacy_notice_query,
        {
            "privacy_notice_id": privacy_notice_id,
            "new_version": int(current_version) + 1,
        },
    )


def create_corresponding_historical_record(
    bind: Connection, privacy_notice_id: str
) -> None:
    """
    Create new historical privacynoticehistory record to match the changes to the privacynotice
    """
    new_version_query: TextClause = text(
        """
        INSERT INTO privacynoticehistory
                (id,
                 privacy_notice_id,
                 name,
                 description,
                 origin,
                 regions,
                 consent_mechanism,
                 data_uses,
                 version,
                 disabled,
                 displayed_in_privacy_center,
                 enforcement_level,
                 has_gpc_flag,
                 internal_description,
                 displayed_in_overlay,
                 displayed_in_api,
                 notice_key)
        SELECT  :new_id,
                :privacy_notice_id,
                name,
                description,
                origin,
                regions,
                consent_mechanism,
                data_uses,
                version,
                disabled,
                displayed_in_privacy_center,
                enforcement_level,
                has_gpc_flag,
                internal_description,
                displayed_in_overlay,
                displayed_in_api,
                notice_key
        FROM   privacynotice
        WHERE  id = :privacy_notice_id 
        """
    )

    bind.execute(
        new_version_query,
        {"privacy_notice_id": privacy_notice_id, "new_id": f"pri_{uuid.uuid4()}"},
    )


def upgrade():
    """
    This data migration disables and bumps the version of enabled Privacy Notices that are not applicable for a
    current system.

    A new Privacy Notice History record is created to have a snapshot of the changes made here.

    This migration gets notices in a better state for new UI.  We originally shipped with out of the box notices
    that were enabled by default, but new UI prevents you from enabling a notice unless it is applicable for a system.

    """
    bind: Connection = op.get_bind()

    enabled_notices: ResultProxy = bind.execute(
        text("SELECT id, data_uses, version FROM privacynotice WHERE disabled = false;")
    )

    expanded_system_data_uses: Set = get_expanded_system_data_uses(bind)

    for privacy_notice_row in enabled_notices:
        notice_data_uses: set[str] = set(privacy_notice_row["data_uses"] or [])
        systems_applicable: bool = bool(
            notice_data_uses.intersection(expanded_system_data_uses)
        )
        if systems_applicable:
            # The enabled privacy notice is relevant for an active system, so no action needed here.
            continue

        # This mimics what happens in the application.  Making a change to a privacy notice (here, disabling it),
        # increases its version number, and creates a corresponding historical record.
        disable_privacy_notice_and_bump_version(
            bind, privacy_notice_row["id"], privacy_notice_row["version"]
        )
        create_corresponding_historical_record(bind, privacy_notice_row["id"])


def downgrade():
    """No downgrade."""



"""
Staged resource ancestor link data migration.

Indexes and constraints are created on this table _after_ data is populated.
If > 1,000,000 stagedresourceancestor records are created as part
of the data migration, the migration will skip creating the indexes
and constraints, and will instead log a message
indicating that index migration should be performed as part of app runtime.

The data migration has the following steps:
- Query for all staged resources and their children
- Build a list in-memory of all ancestor-descendant pairs to insert by recursively
    processing each resource's children
- Write the ancestor links to a CSV string in-memory
- Copy the CSV string into the stagedresourceancestor table via a PostgreSQL COPY
    command
- Create the indexes and constraints on the stagedresourceancestor table if
    < 1,000,000 ancestor links are created


Revision ID: bf713b5a021d
Revises: 5474a47c77da
Create Date: 2025-06-03 09:44:58.769535

"""

import csv
import uuid
from io import StringIO
from pathlib import Path

import sqlalchemy as sa
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
revision = "bf713b5a021d"
down_revision = "5474a47c77da"

branch_labels = None
depends_on = None


def upgrade():
    logger.info("Populating staged resource ancestor links...")

    conn = op.get_bind()

    # Get all resources and their children in batches
    BATCH_SIZE = 500000

    # Query resources in batches using yield_per
    resources_query = sa.text(
        """
        SELECT urn, children
        FROM stagedresource
    """
    )
    resource_children = {}

    # process resources and populate resource_children map in batches
    # to limit memory usage
    for batch in (
        conn.execution_options(stream_results=True)
        .execute(resources_query)
        .yield_per(BATCH_SIZE)
        .partitions()
    ):
        logger.info(f"Processing batch of {BATCH_SIZE} resources")

        # Build resource -> children map for this batch
        for result in batch:
            urn = result.urn
            children = result.children
            if children:
                resource_children[urn] = children
            else:
                # even if no children, we still add the resource to the map
                # so that we maintain a record of all resources and can
                # check for orphaned descendant records later
                resource_children[urn] = []

    # Build list of ancestor-descendant pairs
    ancestor_links = []

    def process_children(ancestor_urn, children, visited=None):
        """Recursively process children and collect ancestor links"""
        if visited is None:
            visited = set()

        for child_urn in children:
            if child_urn not in visited:
                visited.add(child_urn)
                if child_urn not in resource_children:
                    logger.warning(
                        f"Found orphaned descendant URN: {child_urn}, not adding ancestor link"
                    )
                else:
                    # Add direct ancestor link
                    ancestor_links.append(
                        {
                            "id": f"srl_{uuid.uuid4()}",
                            "ancestor_urn": ancestor_urn,
                            "descendant_urn": child_urn,
                        }
                    )

                    # Recursively process this child's children
                    process_children(
                        ancestor_urn, resource_children[child_urn], visited
                    )

    logger.info(
        f"Recursively processing {len(resource_children)} resources for ancestor links in current batch"
    )

    # Process each resource's children recursively
    for ancestor_urn, children in resource_children.items():
        process_children(ancestor_urn, children)

    # remove the resource_children map from memory
    del resource_children

    ancestor_links_count = len(ancestor_links)
    logger.info(f"Found {ancestor_links_count} ancestor links in current batch")

    if ancestor_links_count > 0:
        # Create temporary CSV file
        temp_csv_path = Path("staged_resource_ancestors.csv")
        with open(temp_csv_path, "w", newline="") as csv_file:
            writer = csv.DictWriter(
                csv_file, fieldnames=["id", "ancestor_urn", "descendant_urn"]
            )
            writer.writeheader()
            logger.info(f"Writing {ancestor_links_count} ancestor links to CSV file")
            writer.writerows(ancestor_links)

        del ancestor_links

        # Copy all data from CSV file into table
        logger.info(
            f"Copying {ancestor_links_count} ancestor links from CSV file into stagedresourceancestor table..."
        )
        with open(temp_csv_path, "r") as csv_file:
            copy_query = """
                COPY stagedresourceancestor (id, ancestor_urn, descendant_urn)
                FROM STDIN
                WITH (FORMAT CSV, HEADER TRUE)
            """
            conn.connection.cursor().copy_expert(copy_query, csv_file)

        # Clean up temp file
        temp_csv_path.unlink()

        logger.info(
            f"Completed copying all ancestor links. Total ancestor links created: {ancestor_links_count}"
        )

    if ancestor_links_count < 1000000:

        logger.info("Creating primary key index on stagedresourceancestor table...")

        op.create_index(
            "ix_staged_resource_ancestor_pkey",
            "stagedresourceancestor",
            ["id"],
            unique=True,
        )

        logger.info("Completed creating primary key index")

        logger.info("Creating foreign key constraints stagedresourceancestor table...")

        op.create_foreign_key(
            "fk_staged_resource_ancestor_ancestor",
            "stagedresourceancestor",
            "stagedresource",
            ["ancestor_urn"],
            ["urn"],
            ondelete="CASCADE",
        )

        op.create_foreign_key(
            "fk_staged_resource_ancestor_descendant",
            "stagedresourceancestor",
            "stagedresource",
            ["descendant_urn"],
            ["urn"],
            ondelete="CASCADE",
        )

        logger.info("Completed creating foreign key constraints")

        logger.info("Creating unique constraint on stagedresourceancestor table...")

        op.create_unique_constraint(
            "uq_staged_resource_ancestor",
            "stagedresourceancestor",
            ["ancestor_urn", "descendant_urn"],
        )

        logger.info("Completed creating unique constraint")

        logger.info("Creating indexes on stagedresourceancestor table...")

        op.create_index(
            "ix_staged_resource_ancestor_ancestor",
            "stagedresourceancestor",
            ["ancestor_urn"],
            unique=False,
        )
        op.create_index(
            "ix_staged_resource_ancestor_descendant",
            "stagedresourceancestor",
            ["descendant_urn"],
            unique=False,
        )

        logger.info("Completed creating indexes on stagedresourceancestor table")
    else:
        logger.info(
            "Skipping creation of primary key index, foreign key constraints, unique constraint, and indexes on stagedresourceancestor table because there are more than 1,000,000 ancestor links. These will be populated as part of the `post_upgrade_index_creation.py` task kicked off during application startup."
        )


def downgrade():
    logger.info(
        "Downgrading staged resource ancestor link data migration, populating child_diff_statuses..."
    )

    # Get child diff statuses for each ancestor
    conn = op.get_bind()
    child_diff_statuses_query = """
        UPDATE stagedresource
        SET child_diff_statuses = child_statuses.status_map
        FROM (
            SELECT
                stagedresourceancestor.ancestor_urn,
                jsonb_object_agg(distinct(stagedresource.diff_status), True) as status_map
            FROM stagedresourceancestor
            JOIN stagedresource ON stagedresourceancestor.descendant_urn = stagedresource.urn
            GROUP BY stagedresourceancestor.ancestor_urn
        ) AS child_statuses
        WHERE stagedresource.urn = child_statuses.ancestor_urn
    """
    result = conn.execute(child_diff_statuses_query)
    updated_rows = result.rowcount

    logger.info(
        f"Downgraded staged resource ancestor link data migration, completed populating child_diff_statuses for {updated_rows} rows"
    )

    # Intentionally not dropping the stagedresourceancestor indexes here because they
    # may not have been created as part of the data migration above. If they were not created,
    # then the downgrade would fail trying to drop the non-existent indexes.

    # The indexes and constraints will be dropped as part of the overall table drop that's done
    # as part of the downgrade in the `5474a47c77da_create_staged_resource_ancestor_link_table.py`
    # migration.
